package com.publicobject.issuesbrowser;

import ca.odell.glazedlists.TextFilterable;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

/**
 * An issue models a work effort either due to an existing problem or a desired
 * enhancement.
 *
 * @author James Lemieux
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public class Issue implements TextFilterable, Comparable {
    private Project owner;
    private Integer id = null;
    private Integer statusCode = null;
    private String status = null;
    private Priority priority = null;
    private String resolution = null;
    private String component = null;
    private String version = null;
    private String repPlatform = null;
    private String assignedTo = null;
    private Date deltaTimestamp = null;
    private String subcomponent = null;
    private String reporter = null;
    private String targetMilestone = null;
    private String issueType = null;
    private Date creationTimestamp = null;
    private String qaContact = null;
    private String statusWhiteboard = null;
    private String fileLocation = null;
    private String votes = null;
    private String operatingSystem = null;
    private String shortDescription = null;
    private PeerIssue isDuplicate = null;
    private List keywords = new ArrayList();
    private List blocks = new ArrayList();
    private List cc = new ArrayList();
    private List descriptions = new ArrayList();
    private List attachments = new ArrayList();
    private List activities = new ArrayList();
    private List duplicates = new ArrayList();
    private List dependsOn = new ArrayList();
    private List allUsers = null;
    
    /**
     * Gets all users related to this issue.
     */
    public List getAllUsers() {
        if (allUsers == null) {
            allUsers = new ArrayList();
            if (assignedTo != null) allUsers.add(assignedTo);
            if (reporter != null) allUsers.add(reporter);
            if (qaContact != null) allUsers.add(qaContact);
            for (Iterator d = descriptions.iterator(); d.hasNext(); ) {
                allUsers.add(((Description)d.next()).getWho());
            }
        }
        return allUsers;
    }
    
    /**
     * Creates a new empty issue.
     */
    public Issue(Project owner) {
        super();
        this.owner = owner;
    }
    
    /**
     * Creates a new issue that uses the specified issue as a template.
     */
    public Issue(Project owner, Issue template) {
        super();
        this.owner = owner;
        id = template.id;
        statusCode = template.statusCode;
        status = template.status;
        priority = template.priority;
        resolution = template.resolution;
        component = template.component;
        version = template.version;
        repPlatform = template.repPlatform;
        assignedTo = template.assignedTo;
        deltaTimestamp = template.deltaTimestamp;
        subcomponent = template.subcomponent;
        reporter = template.reporter;
        targetMilestone = template.targetMilestone;
        issueType = template.issueType;
        creationTimestamp = template.creationTimestamp;
        qaContact = template.qaContact;
        statusWhiteboard = template.statusWhiteboard;
        votes = template.votes;
        operatingSystem = template.operatingSystem;
        shortDescription = template.shortDescription;
        keywords.addAll(template.keywords);
        blocks.addAll(template.blocks);
        dependsOn.addAll(template.dependsOn);
        cc.addAll(template.cc);
        descriptions.addAll(template.descriptions);
        attachments.addAll(template.attachments);
        activities.addAll(template.activities);
    }
    
    /**
     * ID of this issue (unique key).
     */
    public Integer getId() {
        return id;
    }
    
    public void setId(Integer id) {
        this.id = id;
    }
    
    /**
     * Get the web address of this issue for use with a browser like IE or Firefox.
     */
    public URL getURL() {
        try {
            return new URL(owner.getBaseUri() + "/issues/show_bug.cgi?id=" + getId());
        } catch (MalformedURLException e) {
            throw new RuntimeException(e);
        }
    }
    
    /**
     * Status code of this issue (load status).
     */
    public Integer getStatusCode() {
        return statusCode;
    }
    
    public void setStatusCode(Integer statusCode) {
        this.statusCode = statusCode;
    }
    
    /**
     * Current status of this issue.
     */
    public String getStatus() {
        return status;
    }
    
    public void setStatus(String status) {
        this.status = status;
    }
    
    /**
     * Priority (severity) assigned to issue.
     */
    public Priority getPriority() {
        return priority;
    }
    
    public void setPriority(Priority priority) {
        this.priority = priority;
    }
    
    /**
     * The issue's resolution, if any
     */
    public String getResolution() {
        return resolution;
    }
    
    public void setResolution(String resolution) {
        this.resolution = resolution;
    }
    
    /**
     * Product against which issue is reported.
     */
    public String getComponent() {
        return component;
    }
    
    public void setComponent(String component) {
        this.component = component;
    }
    
    /**
     * Version associated with component.
     */
    public String getVersion() {
        return version;
    }
    
    public void setVersion(String version) {
        this.version = version;
    }
    
    /**
     * Platform that the issue was reported against.
     */
    public String getRepPlatform() {
        return repPlatform;
    }
    
    public void setRepPlatform(String repPlatform) {
        this.repPlatform = repPlatform;
    }
    
    /**
     * Email of person issue currently assigned to.
     */
    public String getAssignedTo() {
        return assignedTo;
    }
    
    public void setAssignedTo(String assignedTo) {
        this.assignedTo = assignedTo;
    }
    
    /**
     * Last modified timestamp ('yyyy-mm-dd hh:mm:ss').
     */
    public Date getDeltaTimestamp() {
        return deltaTimestamp;
    }
    
    public void setDeltaTimestamp(Date deltaTimestamp) {
        this.deltaTimestamp = deltaTimestamp;
    }
    
    /**
     * Component of component issue reported against.
     */
    public String getSubcomponent() {
        return subcomponent;
    }
    
    public void setSubcomponent(String subcomponent) {
        this.subcomponent = subcomponent;
    }
    
    /**
     * Email of initial issue reporter.
     */
    public String getReporter() {
        return reporter;
    }
    
    public void setReporter(String reporter) {
        this.reporter = reporter;
    }
    
    /**
     * Milestone for this issue's resolution.
     */
    public String getTargetMilestone() {
        return targetMilestone;
    }
    
    public void setTargetMilestone(String targetMilestone) {
        this.targetMilestone = targetMilestone;
    }
    
    /**
     * Nature of issue.  This refers to whether the issue is a defect, task,
     * enhancement, etc.
     */
    public String getIssueType() {
        return issueType;
    }
    
    public void setIssueType(String issueType) {
        this.issueType = issueType;
    }
    
    /**
     * Issue creation timestamp ('yyyy-mm-dd hh:mm:ss').
     */
    public Date getCreationTimestamp() {
        return creationTimestamp;
    }
    
    public void setCreationTimestamp(Date creationTimestamp) {
        this.creationTimestamp = creationTimestamp;
    }
    
    /**
     * Email of the QA contact for this issue.
     */
    public String getQAContact() {
        return qaContact;
    }
    
    public void setQAContact(String qaContact) {
        this.qaContact = qaContact;
    }
    
    /**
     * Free text 'whiteboard' for issue comments.
     */
    public String getStatusWhiteboard() {
        return statusWhiteboard;
    }
    
    public void setStatusWhiteboard(String statusWhiteboard) {
        this.statusWhiteboard = statusWhiteboard;
    }
    
    /**
     * current votes for issue.
     */
    public String getVotes() {
        return votes;
    }
    
    public void setVotes(String votes) {
        this.votes = votes;
    }
    
    /**
     * URL related to issue
     */
    public String getFileLocation() {
        return fileLocation;
    }
    
    public void setFileLocation(String fileLocation) {
        this.fileLocation = fileLocation;
    }
    
    /**
     * Operating system issue reported against.
     */
    public String getOperatingSystem() {
        return operatingSystem;
    }
    
    public void setOperatingSystem(String operatingSystem) {
        this.operatingSystem = operatingSystem;
    }
    
    /**
     * Short description of issue.
     */
    public String getShortDescription() {
        return shortDescription;
    }
    
    public void setShortDescription(String shortDescription) {
        this.shortDescription = shortDescription;
    }
    
    /**
     * List of keywords for this issue.
     */
    public List getKeywords() {
        return keywords;
    }
    
    /**
     * List of email addresses of interested parties.
     */
    public List getCC() {
        return cc;
    }
    
    /**
     * Data from the longdescs table for this issue id. Essentially
     * the log of additional comments.
     */
    public List getDescriptions() {
        return descriptions;
    }
    
    /**
     * Get the attachments to this issue.
     */
    public List getAttachments() {
        return attachments;
    }
    
    /**
     * Get the activity upon this issue.
     */
    public List getActivities() {
        return activities;
    }
    
    /**
     * Other issues which were closed as a duplicate of this issue.
     */
    public List getDuplicates() {
        return duplicates;
    }
    
    /**
     * List of local issue IDs that depend on this one.
     */
    public List getDependsOn() {
        return dependsOn;
    }
    
    /**
     * List of local issue IDs blocked by this one.
     */
    public List getBlocks() {
        return blocks;
    }
    
    /**
     * The issue which this issue was closed as a duplicate of.
     */
    public PeerIssue getDuplicate() {
        return isDuplicate;
    }
    
    public void setDuplicate(PeerIssue isDuplicate) {
        this.isDuplicate = isDuplicate;
    }
    
    /**
     * Write this issue for debugging.
     */
    public String toString() {
        return "Issue " + id + ": " + getPriority().getRating();
    }
    
    /**
     * Compares two issues by ID.
     */
    public int compareTo(Object other) {
        if (other == null) return -1;
        Issue otherIssue = (Issue)(Issue)other;
        return id.compareTo(otherIssue.id);
    }
    
    /**
     * Gets the strings to filter this issue by.
     */
    public void getFilterStrings(List baseList) {
        baseList.add(id.toString());
        baseList.add(issueType);
        baseList.add(priority.toString());
        baseList.add(status);
        baseList.add(resolution);
        baseList.add(shortDescription);
        baseList.add(component);
        baseList.add(subcomponent);
        for (Iterator d = descriptions.iterator(); d.hasNext(); ) {
            ((Description)d.next()).getFilterStrings(baseList);
        }
    }
}
