package com.publicobject.issuesbrowser;

import ca.odell.glazedlists.EventList;
import ca.odell.glazedlists.GlazedLists;
import java.io.IOException;
import java.net.NoRouteToHostException;
import java.net.UnknownHostException;
import java.security.AccessControlException;

/**
 * This loads issues by project as they are requested. When a new project is
 * requested, a working project may be interrupted. This may have violent side
 * effects such as InterruptedExceptions printed to the console by certain
 * XML parsing libraries that aren't exactly interruption friendly.
 *
 * <p>Issues are streamed to the issues list as they are loaded.
 *
 * @author <a href="jesse@odel.on.ca">Jesse Wilson</a>
 */
public class IssueLoader implements Runnable {
    private Project project = null;
    private Throbber throbber = null;
    private Thread issueLoaderThread = null;
    private EventList issuesList = null;
    
    public IssueLoader(EventList issuesList, Throbber throbber) {
        super();
        this.issuesList = GlazedLists.threadSafeList(issuesList);
        this.throbber = throbber;
    }
    
    public void setProject(Project project) {
        synchronized (this) {
            this.project = project;
            issueLoaderThread.interrupt();
            notify();
        }
    }
    
    public void start() {
        issueLoaderThread = new Thread(this);
        issueLoaderThread.setPriority(Thread.NORM_PRIORITY);
        issueLoaderThread.start();
    }
    
    public void run() {
        Project currentProject = null;
        while (true) {
            try {
                synchronized (this) {
                    if (project == null) wait();
                    Thread.interrupted();
                    if (project == null) continue;
                    currentProject = project;
                    project = null;
                }
                throbber.setOn();
                issuesList.clear();
                IssuezillaXMLParser.loadIssues(issuesList, currentProject.getXMLUri(), currentProject);
            } catch (UnknownHostException e) {
                Exceptions.getInstance().handle(e);
            } catch (NoRouteToHostException e) {
                Exceptions.getInstance().handle(e);
            } catch (AccessControlException e) {
                Exceptions.getInstance().handle(e);
            } catch (IOException e) {
                if (e.getCause() instanceof InterruptedException) {
                } else if (e.getMessage().equals("Parsing failed java.lang.InterruptedException")) {
                } else {
                    e.printStackTrace();
                }
            } catch (RuntimeException e) {
                if (e.getCause() instanceof InterruptedException) {
                } else if (e.getCause() instanceof IOException && e.getCause().getMessage().equals("Parsing failed Lock interrupted")) {
                } else {
                    throw e;
                }
            } catch (InterruptedException e) {
            } finally {
                throbber.setOff();
            }
        }
    }
}
