package com.publicobject.issuesbrowser;

import ca.odell.glazedlists.CollectionList;
import ca.odell.glazedlists.EventList;
import ca.odell.glazedlists.UniqueList;
import ca.odell.glazedlists.event.ListEvent;
import ca.odell.glazedlists.event.ListEventListener;
import ca.odell.glazedlists.matchers.AbstractMatcherEditor;
import ca.odell.glazedlists.matchers.Matcher;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * An UsersMatcherEditor is a matcher editor that filters based on the selected
 * users.
 *
 * @author <a href="mailto:kevin@swank.ca">Kevin Maltby</a>
 * @author <a href="mailto:jesse@swank.ca">Jesse Wilson</a>
 */
public class UsersMatcherEditor extends AbstractMatcherEditor {
    
    /**
     * a list of users 
     */
    private EventList allUsers = null;
    
    /**
     * a list that maintains selection 
     */
    private EventList selectedUsers = null;
    
    /**
     * Create a filter list that filters the specified source list, which
     * must contain only Issue objects.
     */
    public UsersMatcherEditor(EventList source) {
        super();
        allUsers = new UniqueList(new CollectionList(source, new IssueUserator()));
    }
    
    /**
     * Sets the selection driven EventList which triggers filter changes.
     */
    public void setSelectionList(EventList usersSelectedList) {
        this.selectedUsers = usersSelectedList;
        usersSelectedList.addListEventListener(new SelectionChangeEventList());
    }
    
    /**
     * Allow access to the unique list of users
     */
    public EventList getUsersList() {
        return allUsers;
    }
    
    /**
     * An EventList to respond to changes in selection from the ListEventViewer.
     */
    private final class SelectionChangeEventList implements ListEventListener {
        
        private SelectionChangeEventList() {
            super();
        }
        
        /**
         * {@inheritDoc} 
         */
        public void listChanged(ListEvent listChanges) {
            if (selectedUsers.isEmpty() || selectedUsers.size() == allUsers.size()) {
                fireMatchAll();
                return;
            }
            final UserMatcher newUserMatcher = new UserMatcher(selectedUsers);
            final Matcher previousMatcher = getMatcher();
            if (!(previousMatcher instanceof UserMatcher)) {
                fireConstrained(newUserMatcher);
                return;
            }
            final UserMatcher previousUserMatcher = (UserMatcher)(UsersMatcherEditor.UserMatcher)previousMatcher;
            boolean relaxed = newUserMatcher.isRelaxationOf(previousMatcher);
            boolean constrained = previousUserMatcher.isRelaxationOf(newUserMatcher);
            if (relaxed && constrained) return;
            if (relaxed) fireRelaxed(newUserMatcher); else if (constrained) fireConstrained(newUserMatcher); else fireChanged(newUserMatcher);
        }
    }
    
    /**
     * If the set of users for this matcher contains any of an
     * issue's users, that issue matches.
     */
    private static class UserMatcher implements Matcher {
        private final Set users;
        
        /**
         * Create a new {@link UserMatcher}, creating a private copy
         * of the specified {@link Collection} to match against. A private
         * copy is made because {@link Matcher}s must be immutable.
         */
        public UserMatcher(Collection users) {
            super();
            this.users = new HashSet(users);
        }
        
        /**
         * @return true if this matches every {@link Issue} the other matches.
         */
        public boolean isRelaxationOf(Matcher other) {
            if (!(other instanceof UserMatcher)) return false;
            UserMatcher otherUserMatcher = (UserMatcher)(UsersMatcherEditor.UserMatcher)other;
            return users.containsAll(otherUserMatcher.users);
        }
        
        /**
         * Test whether to include or not include the specified issue based
         * on whether or not their user is selected.
         */
        public boolean matches(Object x0) {
            Issue issue = (Issue)x0;
            for (Iterator i = issue.getAllUsers().iterator(); i.hasNext(); ) {
                if (this.users.contains(i.next())) return true;
            }
            return false;
        }
        /*missing*/
    }
}
