package ca.odell.glazedlists;

import java.util.*;
import ca.odell.glazedlists.event.*;
import ca.odell.glazedlists.impl.adt.*;

/**
 * A list that acts like a tree in that it contains child elements to nodes contained in
 * another list. An example usage would be to wrap a parent list containing record albums
 * and use the CollectionList to display the songs on the album.
 * 
 * <p>The actual mapping from the parent list to the child list (record to songs in the
 * above example) is done by a {@link CollectionList.Model} that is provided to the
 * constructor.
 *
 * <p><table border="1" width="100%" cellpadding="3" cellspacing="0">
 * <tr class="tableheadingcolor"><td colspan=2><font size="+2"><b>EventList Overview</b></font></td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Writable:</b></td><td>only {@link #set(int,Object)} and {@link #remove(int)}</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Concurrency:</b></td><td>thread ready, not thread safe</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Performance:</b></td><td>reads: O(log N), writes O(log N)</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Memory:</b></td><td>96 bytes per element</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Unit Tests:</b></td><td>N/A</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Issues:</b></td><td>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=96">96</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=162">162</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=257">257</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=265">265</a>
 * </td></tr>
 * </table>
 *
 * @see CollectionList.Model
 *
 * @author <a href="mailto:rob@starlight-systems.com">Rob Eden</a>
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public class CollectionList extends TransformedList implements ListEventListener {
    
    /**
     * This is a hack - we need a temporary value when inserting into IndexedTrees, and this is the one we use. 
     */
    private final ChildElement EMPTY_CHILD_ELEMENT = new SimpleChildElement(Collections.EMPTY_LIST, null);
    
    /**
     * used to extract children 
     */
    private final Model model;
    
    /**
     * Barcode containing the node mappings. There is a black node for each parent
     * followed by a white node for each of its children.
     */
    private final Barcode barcode = new Barcode();
    
    /**
     * the Lists and EventLists that this is composed of 
     */
    private final IndexedTree childElements = new IndexedTree();
    
    /**
     * Create a {@link CollectionList} that's contents will be the children of the
     * elements in the specified source {@link EventList}.
     */
    public CollectionList(EventList source, Model model) {
        super(source);
        if (model == null) throw new IllegalArgumentException("Collection map cannot be null");
        this.model = model;
        for (int i = 0; i < source.size(); i++) {
            List children = model.getChildren(source.get(i));
            IndexedTreeNode node = childElements.addByNode(i, EMPTY_CHILD_ELEMENT);
            node.setValue(createChildElementForList(children, node));
            barcode.addBlack(barcode.size(), 1);
            if (!children.isEmpty()) barcode.addWhite(barcode.size(), children.size());
        }
        source.addListEventListener(this);
    }
    
    /**
     * {@inheritDoc} 
     */
    public int size() {
        return barcode.whiteSize();
    }
    
    /**
     * {@inheritDoc} 
     */
    public Object get(int index) {
        ChildElement childElement = getChildElement(index);
        int childIndexInParent = barcode.getWhiteSequenceIndex(index);
        return childElement.get(childIndexInParent);
    }
    
    /**
     * {@inheritDoc} 
     */
    public Object set(int index, Object value) {
        ChildElement childElement = getChildElement(index);
        int childIndexInParent = barcode.getWhiteSequenceIndex(index);
        return childElement.set(childIndexInParent, value);
    }
    
    /**
     * {@inheritDoc} 
     */
    public Object remove(int index) {
        ChildElement childElement = getChildElement(index);
        int childIndexInParent = barcode.getWhiteSequenceIndex(index);
        return childElement.remove(childIndexInParent);
    }
    
    /**
     * Return the index of the first child in the CollectionList for the given parent
     * index. This can be very useful for things like selecting the children in a
     * CollectionList when the parent is selected in another list.
     *
     * @see #childEndingIndex
     */
    public int childStartingIndex(int parentIndex) {
        if (parentIndex < 0) throw new IndexOutOfBoundsException("Invalid index: " + parentIndex);
        if (parentIndex >= source.size()) throw new IndexOutOfBoundsException("Invalid index: " + parentIndex);
        int parentFullIndex = barcode.getIndex(parentIndex, Barcode.BLACK);
        int childFullIndex = parentFullIndex + 1;
        if (childFullIndex >= barcode.size()) return -1;
        if (barcode.get(childFullIndex) != Barcode.WHITE) return -1;
        int childIndex = childFullIndex - (parentIndex + 1);
        assert (barcode.getWhiteIndex(childFullIndex) == childIndex);
        return childIndex;
    }
    
    /**
     * Return the index of the last child in the CollectionList for the given parent
     * index. This can be very useful for things like selecting the children in a
     * CollectionList when the parent is selected in another list.
     *
     * @see #childStartingIndex
     */
    public int childEndingIndex(int parentIndex) {
        if (parentIndex < 0) throw new IndexOutOfBoundsException("Invalid index: " + parentIndex);
        if (parentIndex >= source.size()) throw new IndexOutOfBoundsException("Invalid index: " + parentIndex);
        int nextParentFullIndex = (parentIndex == barcode.blackSize() - 1) ? barcode.size() : barcode.getIndex(parentIndex + 1, Barcode.BLACK);
        int lastWhiteBeforeNextParent = nextParentFullIndex - 1;
        if (barcode.get(lastWhiteBeforeNextParent) == Barcode.BLACK) return -1;
        int childIndex = lastWhiteBeforeNextParent - (parentIndex + 1);
        assert (barcode.getWhiteIndex(lastWhiteBeforeNextParent) == childIndex);
        return childIndex;
    }
    
    /**
     * Handle changes in the parent list. We'll need to update our node list sizes.
     */
    public void listChanged(ListEvent listChanges) {
        updates.beginEvent();
        while (listChanges.next()) {
            int index = listChanges.getIndex();
            int type = listChanges.getType();
            if (type == ListEvent.INSERT) {
                handleInsert(index);
            } else if (type == ListEvent.DELETE) {
                handleDelete(index);
            } else if (type == ListEvent.UPDATE) {
                handleDelete(index);
                handleInsert(index);
            }
        }
        updates.commitEvent();
    }
    
    /**
     * Helper for {@link #listChanged(ListEvent)} when inserting.
     */
    private void handleInsert(int parentIndex) {
        int absoluteIndex = getAbsoluteIndex(parentIndex);
        Object parent = source.get(parentIndex);
        List children = model.getChildren(parent);
        IndexedTreeNode node = childElements.addByNode(parentIndex, EMPTY_CHILD_ELEMENT);
        node.setValue(createChildElementForList(children, node));
        barcode.addBlack(absoluteIndex, 1);
        if (!children.isEmpty()) barcode.addWhite(absoluteIndex + 1, children.size());
        if (!children.isEmpty()) {
            int childIndex = absoluteIndex - parentIndex;
            updates.addInsert(childIndex, childIndex + children.size() - 1);
        }
    }
    
    /**
     * Helper for {@link #listChanged(ListEvent)} when deleting.
     */
    private void handleDelete(int sourceIndex) {
        int parentIndex = getAbsoluteIndex(sourceIndex);
        int nextParentIndex = getAbsoluteIndex(sourceIndex + 1);
        ChildElement removedChildElement = (CollectionList.ChildElement)childElements.removeByIndex(sourceIndex).getValue();
        removedChildElement.dispose();
        int childCount = nextParentIndex - parentIndex - 1;
        barcode.remove(parentIndex, 1 + childCount);
        if (childCount > 0) {
            int firstDeletedChildIndex = parentIndex - sourceIndex;
            int firstNotDeletedChildIndex = firstDeletedChildIndex + childCount;
            updates.addDelete(firstDeletedChildIndex, firstNotDeletedChildIndex - 1);
        }
    }
    
    /**
     * Get the child element for the specified child index.
     */
    private ChildElement getChildElement(int childIndex) {
        if (childIndex < 0) throw new IndexOutOfBoundsException("Invalid index: " + childIndex);
        if (childIndex >= size()) throw new IndexOutOfBoundsException("Index: " + childIndex + ", Size: " + size());
        int parentIndex = barcode.getBlackBeforeWhite(childIndex);
        return (CollectionList.ChildElement)childElements.getNode(parentIndex).getValue();
    }
    
    /**
     * Create a {@link ChildElement} for the specified List.
     */
    private ChildElement createChildElementForList(List children, IndexedTreeNode node) {
        if (children instanceof EventList) return new EventChildElement((EventList)(EventList)children, node); else return new SimpleChildElement(children, node);
    }
    
    /**
     * Get the absolute index for the specified parent index. This may be virtual
     * if the parent index is one greater than the last element. This is useful
     * for calculating the size of a range by using the location of its follower.
     */
    private int getAbsoluteIndex(int parentIndex) {
        if (parentIndex < barcode.blackSize()) {
            return barcode.getIndex(parentIndex, Barcode.BLACK);
        } else if (parentIndex == barcode.blackSize()) {
            return barcode.size();
        } else {
            throw new IndexOutOfBoundsException();
        }
    }
    
    /**
     * Models a list held by the CollectionList.
     */
    private interface ChildElement {
        
        public Object get(int index);
        
        public Object remove(int index);
        
        public Object set(int index, Object element);
        
        public void dispose();
    }
    
    /**
     * Provides the logic to map a parent record (e.g., a records album) to its children
     * (e.g., the songs on the record). Serves basically the same purpose as
     * {@link javax.swing.tree.TreeModel} does to a JTree in Swing.
     *
     * @see CollectionList
     * @see GlazedLists#listCollectionListModel()
     */
    public interface Model {
        
        /**
         * Return a list of the child nodes for a parent node.
         *
         * @param parent The parent node.
         * @return A List containing the child nodes.
         */
        public List getChildren(Object parent);
    }
    
    /**
     * Manages a standard List that does not implement {@link EventList}.
     */
    private class SimpleChildElement implements ChildElement {
        private List children;
        private IndexedTreeNode node;
        
        public SimpleChildElement(List children, IndexedTreeNode node) {
            super();
            this.children = children;
            this.node = node;
        }
        
        public Object get(int index) {
            return children.get(index);
        }
        
        public Object remove(int index) {
            Object removed = children.remove(index);
            int parentIndex = node.getIndex();
            int absoluteIndex = getAbsoluteIndex(parentIndex);
            int firstChildIndex = absoluteIndex + 1;
            barcode.remove(firstChildIndex + index, 1);
            int childOffset = absoluteIndex - parentIndex;
            updates.beginEvent();
            updates.addDelete(index + childOffset);
            updates.commitEvent();
            return removed;
        }
        
        public Object set(int index, Object element) {
            Object replaced = children.set(index, element);
            int parentIndex = node.getIndex();
            int absoluteIndex = getAbsoluteIndex(parentIndex);
            int childOffset = absoluteIndex - parentIndex;
            updates.beginEvent();
            updates.addUpdate(index + childOffset);
            updates.commitEvent();
            return replaced;
        }
        
        public void dispose() {
        }
    }
    
    /**
     * Monitors changes to a member EventList and forwards changes to all listeners
     * of the CollectionList.
     */
    private class EventChildElement implements ChildElement, ListEventListener {
        private EventList children;
        private IndexedTreeNode node;
        
        public EventChildElement(EventList children, IndexedTreeNode node) {
            super();
            this.children = children;
            this.node = node;
            children.addListEventListener(this);
        }
        
        public Object get(int index) {
            return children.get(index);
        }
        
        public Object remove(int index) {
            return children.remove(index);
        }
        
        public Object set(int index, Object element) {
            return children.set(index, element);
        }
        
        public void listChanged(ListEvent listChanges) {
            int parentIndex = node.getIndex();
            int absoluteIndex = getAbsoluteIndex(parentIndex);
            int nextNodeIndex = getAbsoluteIndex(parentIndex + 1);
            int firstChildIndex = absoluteIndex + 1;
            int previousChildrenCount = nextNodeIndex - firstChildIndex;
            if (previousChildrenCount > 0) barcode.remove(firstChildIndex, previousChildrenCount);
            if (!children.isEmpty()) barcode.addWhite(firstChildIndex, children.size());
            int childOffset = absoluteIndex - parentIndex;
            updates.beginEvent();
            while (listChanges.next()) {
                int index = listChanges.getIndex();
                int type = listChanges.getType();
                updates.addChange(type, index + childOffset);
            }
            updates.commitEvent();
        }
        
        public void dispose() {
            children.removeListEventListener(this);
        }
    }
}
