package ca.odell.glazedlists;

import ca.odell.glazedlists.event.*;
import ca.odell.glazedlists.impl.adt.*;
import java.util.*;

/**
 * A grouping list contains elements which are themselves Lists. Those Lists
 * are infact elements of the source list which have been grouped together into
 * a List. The logic of how to group the source elements into groups is specified
 * via a Comparator. Elements for which the Comparator returns 0 are guaranteed
 * to be contained within the same group within this GroupingList. This implies
 * that source elements may only participate in a single group within this
 * GroupingList.
 *
 * <p> Further transformations may be layered on top of this GroupingList to
 * transform the group lists into any other desirable form.
 *
 * <p><table border="1" width="100%" cellpadding="3" cellspacing="0">
 * <tr class="tableheadingcolor"><td colspan=2><font size="+2"><b>EventList Overview</b></font></td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Writable:</b></td><td>yes</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Concurrency:</b></td><td>thread ready, not thread safe</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Performance:</b></td><td>reads: O(log N), writes O(log N)</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Memory:</b></td><td></td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Unit Tests:</b></td><td>GroupingListTest</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Issues:</b></td><td>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=281">281</a>
 * </td></tr>
 * </table>
 *
 * @author James Lemieux
 */
public final class GroupingList extends TransformedList {
    
    /**
     * A temporary barcode data structure is created when processing ListEvents in this GroupingList.
     * These constants identify which indexes remain to be processed.
     */
    private static final Object TODO = Barcode.BLACK;
    private static final Object DONE = Barcode.WHITE;
    
    /**
     * The comparator used to determine the groups. 
     */
    private Comparator comparator;
    
    /**
     * The GroupLists defined by the comparator. They are stored in an IndexedTree so their indices can be quickly updated. 
     */
    private IndexedTree groupLists = new IndexedTree();
    
    /**
     * The data structure which tracks which source elements are considered UNIQUE
     * (the first element of a group) and which are DUPLICATE (any group element NOT the first).
     */
    private Barcode barcode = new Barcode();
    
    /**
     * Entries in barcode are one of these constants to indicate if the element at an index is a UNIQUE or DUPLICATE. 
     */
    private static final Object UNIQUE = Barcode.BLACK;
    private static final Object DUPLICATE = Barcode.WHITE;
    
    /**
     * Used only in temporary data structures to flag deleting of the FIRST group element when more elements exist. 
     */
    private static final Object UNIQUE_WITH_DUPLICATE = null;
    
    /**
     * For reporting which side of an element its group is located. 
     */
    private static final int LEFT_GROUP = -1;
    private static final int NO_GROUP = 0;
    private static final int RIGHT_GROUP = 1;
    
    /**
     * Creates a {@link GroupingList} that determines groupings via the
     * {@link Comparable} interface which all elements of the <code>source</code>
     * are assumed to implement.
     */
    public GroupingList(EventList source) {
        this(source, (Comparator)GlazedLists.comparableComparator());
    }
    
    /**
     * Creates a {@link GroupingList} that determines groups using the specified
     * {@link Comparator}.
     *
     * @param source the {@link EventList} containing elements to be grouped
     * @param comparator the {@link Comparator} used to determine groupings
     */
    public GroupingList(EventList source, Comparator comparator) {
        this(new SortedList(source, comparator), comparator);
    }
    
    /**
     * A private constructor which provides a convenient handle to the
     * {@link SortedList} which will serve as the source of this list.
     *
     * @param source the elements to be grouped arranged in sorted order
     * @param comparator the {@link Comparator} used to determine groupings
     */
    private GroupingList(SortedList source, Comparator comparator) {
        super(source);
        this.comparator = comparator;
        this.buildDataStructures();
        source.addListEventListener(this);
    }
    
    /**
     * Populates the internal data structures of GroupingList by examining
     * adjacent entries within the source SortedList to check if they belong
     * to the same group.
     */
    private void buildDataStructures() {
        if (!barcode.isEmpty()) throw new IllegalStateException("Attempted to build internal data structures twice.");
        for (int i = 0; i < source.size(); i++) barcode.add(i, groupTogether(i, i - 1) ? DUPLICATE : UNIQUE, 1);
        for (int i = 0; i < barcode.colourSize(UNIQUE); i++) this.insertGroupList(i);
    }
    
    /**
     * Creates and inserts a new GroupList at the specified <code>index</code>.
     *
     * @param index the location at which to insert an empty GroupList
     */
    private void insertGroupList(int index) {
        final GroupList groupList = new GroupList();
        final IndexedTreeNode indexedTreeNode = this.groupLists.addByNode(index, groupList);
        groupList.setTreeNode(indexedTreeNode);
    }
    
    /**
     * Removes the GroupList at the given <code>index</code>.
     *
     * @param index the location at which to remove a GroupList
     */
    private void removeGroupList(int index) {
        final IndexedTreeNode indexedTreeNode = this.groupLists.removeByIndex(index);
        ((GroupingList.GroupList)indexedTreeNode.getValue()).setTreeNode(null);
    }
    
    /**
     * {@inheritDoc} 
     */
    public int size() {
        return barcode.blackSize();
    }
    
    /**
     * {@inheritDoc} 
     */
    protected int getSourceIndex(int index) {
        return barcode.getIndex(index, UNIQUE);
    }
    
    /**
     * {@inheritDoc} 
     */
    protected boolean isWritable() {
        return true;
    }
    
    /**
     * {@inheritDoc} 
     */
    public void listChanged(ListEvent listChanges) {
        final Barcode toDoList = new Barcode();
        toDoList.addWhite(0, barcode.size());
        final LinkedList removedValues = new LinkedList();
        while (listChanges.next()) {
            final int changeIndex = listChanges.getIndex();
            final int changeType = listChanges.getType();
            if (changeType == ListEvent.INSERT) {
                barcode.add(changeIndex, UNIQUE, 1);
                toDoList.add(changeIndex, TODO, 1);
            } else if (changeType == ListEvent.UPDATE) {
                if (barcode.get(changeIndex) == UNIQUE) {
                    if (changeIndex + 1 < barcode.size() && barcode.get(changeIndex + 1) == DUPLICATE) {
                        barcode.set(changeIndex, UNIQUE, 2);
                        toDoList.set(changeIndex, TODO, 1);
                    }
                }
            } else if (changeType == ListEvent.DELETE) {
                Object deleted = barcode.get(changeIndex);
                barcode.remove(changeIndex, 1);
                toDoList.remove(changeIndex, 1);
                if (deleted == UNIQUE) {
                    if (changeIndex < barcode.size() && barcode.get(changeIndex) == DUPLICATE) {
                        barcode.set(changeIndex, UNIQUE, 1);
                        deleted = UNIQUE_WITH_DUPLICATE;
                    }
                }
                removedValues.addLast(deleted);
            }
        }
        updates.beginEvent(true);
        listChanges.reset();
        while (listChanges.next()) {
            final int changeIndex = listChanges.getIndex();
            final int changeType = listChanges.getType();
            if (changeType == ListEvent.INSERT) {
                if (tryJoinExistingGroup(changeIndex, toDoList) == NO_GROUP) {
                    final int groupIndex = barcode.getColourIndex(changeIndex, UNIQUE);
                    insertGroupList(groupIndex);
                    updates.addInsert(groupIndex);
                } else {
                    updates.addUpdate(barcode.getColourIndex(changeIndex, true, UNIQUE));
                }
            } else if (changeType == ListEvent.UPDATE) {
                int oldGroup = 0;
                if (toDoList.get(changeIndex) == TODO) oldGroup = RIGHT_GROUP; else if (barcode.get(changeIndex) == DUPLICATE) oldGroup = LEFT_GROUP; else if (barcode.get(changeIndex) == UNIQUE) oldGroup = NO_GROUP;
                int newGroup = tryJoinExistingGroup(changeIndex, toDoList);
                int groupIndex = barcode.getColourIndex(changeIndex, true, UNIQUE);
                if (newGroup == NO_GROUP) {
                    if (oldGroup == NO_GROUP) {
                        updates.addUpdate(groupIndex);
                    } else if (oldGroup == LEFT_GROUP) {
                        updates.addUpdate(groupIndex - 1);
                        insertGroupList(groupIndex);
                        updates.addInsert(groupIndex);
                    } else if (oldGroup == RIGHT_GROUP) {
                        insertGroupList(groupIndex);
                        updates.addInsert(groupIndex);
                        updates.addUpdate(groupIndex + 1);
                    }
                } else if (newGroup == LEFT_GROUP) {
                    if (oldGroup == NO_GROUP) {
                        updates.addUpdate(groupIndex);
                        removeGroupList(groupIndex + 1);
                        updates.addDelete(groupIndex + 1);
                    } else if (oldGroup == LEFT_GROUP) {
                        updates.addUpdate(groupIndex);
                    } else if (oldGroup == RIGHT_GROUP) {
                        updates.addUpdate(groupIndex);
                        if (groupIndex + 1 < barcode.blackSize()) updates.addUpdate(groupIndex + 1);
                    }
                } else if (newGroup == RIGHT_GROUP) {
                    if (oldGroup == NO_GROUP) {
                        removeGroupList(groupIndex);
                        updates.addDelete(groupIndex);
                        updates.addUpdate(groupIndex);
                    } else if (oldGroup == LEFT_GROUP) {
                        if (groupIndex - 1 >= 0) updates.addUpdate(groupIndex - 1);
                        updates.addUpdate(groupIndex);
                    } else if (oldGroup == RIGHT_GROUP) {
                        updates.addUpdate(groupIndex);
                    }
                }
            } else if (changeType == ListEvent.DELETE) {
                final Object deleted = removedValues.removeFirst();
                final int sourceDeletedIndex = deleted == DUPLICATE ? changeIndex - 1 : changeIndex;
                final int groupDeletedIndex = sourceDeletedIndex < barcode.size() ? barcode.getBlackIndex(sourceDeletedIndex, true) : barcode.blackSize();
                if (deleted == UNIQUE) {
                    removeGroupList(groupDeletedIndex);
                    updates.addDelete(groupDeletedIndex);
                } else {
                    updates.addUpdate(groupDeletedIndex);
                }
            }
        }
        updates.commitEvent();
    }
    
    /**
     * Handles whether this change index has a neighbour that existed prior
     * to a current change and that the values are equal. This adjusts the
     * duplicates list a found pair of such changes if they exist.
     *
     * @return NO_GROUP if no neighbour was found. In this case the value at the specified
     *      index will be marked as unique, but not temporarily so. Returns LEFT_GROUP if
     *      a neighbour was found on the left, and RIGHT_GROUP if a neighbour was found on the
     *      right. In non-zero cases the duplicates list is updated.
     */
    private int tryJoinExistingGroup(int changeIndex, Barcode toDoList) {
        int predecessorIndex = changeIndex - 1;
        if (groupTogether(predecessorIndex, changeIndex)) {
            barcode.set(changeIndex, DUPLICATE, 1);
            return LEFT_GROUP;
        }
        int successorIndex = changeIndex + 1;
        while (true) {
            if (groupTogether(changeIndex, successorIndex)) {
                if (toDoList.get(successorIndex) == DONE) {
                    barcode.set(changeIndex, UNIQUE, 1);
                    barcode.set(successorIndex, DUPLICATE, 1);
                    return RIGHT_GROUP;
                } else {
                    successorIndex++;
                }
            } else {
                barcode.set(changeIndex, UNIQUE, 1);
                return NO_GROUP;
            }
        }
    }
    
    public Object get(int x0) {
        int index = (int)x0;
        return (List)this.groupLists.get(index);
    }
    
    /**
     * Tests if the specified values should be grouped together.
     *
     * @param sourceIndex0 the first index of the source list to test
     * @param sourceIndex1 the second index of the source list to test
     *
     * @return true iff the values at the specified index are equal according
     *      to the Comparator which defines the grouping logic; false if
     *      either index is out of range or the values shall not be grouped
     */
    private boolean groupTogether(int sourceIndex0, int sourceIndex1) {
        if (sourceIndex0 < 0 || sourceIndex0 >= source.size()) return false;
        if (sourceIndex1 < 0 || sourceIndex1 >= source.size()) return false;
        return comparator.compare(source.get(sourceIndex0), source.get(sourceIndex1)) == 0;
    }
    
    /**
     * {@inheritDoc} 
     */
    public Object remove(int x0) {
        int index = (int)x0;
        if (index < 0 || index >= size()) throw new IndexOutOfBoundsException("Cannot remove at " + index + " on list of size " + size());
        final List removed = (List)this.get(index);
        final List result = new ArrayList(removed);
        removed.clear();
        return result;
    }
    
    /**
     * {@inheritDoc} 
     */
    public Object set(int x0, Object x1) {
        int index = (int)x0;
        List value = (List)x1;
        if (index < 0 || index >= size()) throw new IndexOutOfBoundsException("Cannot set at " + index + " on list of size " + size());
        updates.beginEvent(true);
        final List result = (List)this.remove(index);
        this.add(index, value);
        updates.commitEvent();
        return result;
    }
    
    /**
     * This version of add will distribute all elements within the given
     * <code>value</code> List into groups. Existing groups will be reused and
     * new groups will be created as needed. As such, the <code>index</code>
     * argument is meaningless.
     *
     * <p><strong><font color="#FF0000">Warning:</font></strong> This method
     * breaks the contract required by {@link List#add(int, Object}.
     */
    public void add(int x0, Object x1) {
        int index = (int)x0;
        List value = (List)x1;
        source.addAll(value);
    }
    
    /**
     * {@inheritDoc} 
     */
    public void dispose() {
        ((SortedList)(SortedList)this.source).dispose();
        super.dispose();
    }
    
    /**
     * This is the List implementation used to store groups created by this
     * GroupingList. It defines all mutator methods by mapping them to mutations
     * on the source list of GroupList. Thus, writes to this GroupList effect
     * all Lists sitting under GroupList.
     */
    private class GroupList extends AbstractList {
        
        private GroupList() {
            super();
        }
        
        /**
         * The node within {@link groupLists} that records the index of this GroupList within the GroupingList. 
         */
        private IndexedTreeNode treeNode;
        
        /**
         * Attach the IndexedTreeNode that tracks this GroupLists position to the
         * GroupList itself so it can look up its own position.
         */
        private void setTreeNode(IndexedTreeNode treeNode) {
            this.treeNode = treeNode;
        }
        
        /**
         * Returns the inclusive index of the start of this {@link GroupList}
         * within the source {@link SortedList}.
         */
        private int getStartIndex() {
            final int groupIndex = this.treeNode.getIndex();
            return GroupingList.this.getSourceIndex(groupIndex);
        }
        
        /**
         * Returns the exclusive index of the end of this {@link GroupList}
         * within the source {@link SortedList}.
         */
        private int getEndIndex() {
            final int groupIndex = this.treeNode.getIndex();
            if (groupIndex < barcode.blackSize() - 1) {
                return barcode.getIndex(groupIndex + 1, UNIQUE);
            } else {
                return barcode.size();
            }
        }
        
        private int getSourceIndex(int index) {
            return this.getStartIndex() + index;
        }
        
        /**
         * {@inheritDoc} 
         */
        public Object set(int index, Object element) {
            return source.set(this.getSourceIndex(index), element);
        }
        
        /**
         * {@inheritDoc} 
         */
        public Object get(int index) {
            return source.get(this.getSourceIndex(index));
        }
        
        /**
         * {@inheritDoc} 
         */
        public int size() {
            return this.getEndIndex() - this.getStartIndex();
        }
        
        /**
         * {@inheritDoc} 
         */
        public void clear() {
            source.subList(this.getStartIndex(), this.getEndIndex()).clear();
        }
        
        /**
         * {@inheritDoc} 
         */
        public Object remove(int index) {
            return source.remove(this.getSourceIndex(index));
        }
        
        /**
         * {@inheritDoc} 
         */
        public void add(int index, Object element) {
            source.add(this.getSourceIndex(index), element);
        }
    }
    /*missing*/
    /*missing*/
    /*missing*/
    /*missing*/
}
