package ca.odell.glazedlists;

import ca.odell.glazedlists.event.*;
import ca.odell.glazedlists.impl.adt.*;
import java.util.*;

/**
 * An {@link EventList} to provide index-based selection features.  This
 * {@link EventList} does not perform a transformation on the source, but
 * instead provides two additional {@link EventList}s:
 * <ul>
 *   <li>{@link #getSelected() Selected} - an {@link EventList} that contains only the selected values.</li>
 *   <li>{@link #getDeselected() Deselected} - an {@link EventList} that contains only the deselected values.</li>
 * </ul>
 *
 * <p>This design is intended to allow the sharing of selection logic between
 * both of our supported GUI toolkits as well as being available for use in
 * non-GUI applications and for index-based filtering.
 *
 * <p><strong><font color="#FF0000">Warning:</font></strong> This class is
 * thread ready but not thread safe. See {@link EventList} for an example
 * of thread safe code.
 *
 * @author <a href="mailto:kevin@swank.ca">Kevin Maltby</a>
 */
public class ListSelection {
    
    /**
     * A selection mode where at most one element may be selected at one time.
     * For convenience, this value equals {@link javax.swing.ListSelectionModel#SINGLE_SELECTION}.
     */
    public static final int SINGLE_SELECTION = 0;
    
    /**
     * A selection mode where at most one range of elements may be selected at one time.
     * For convenience, this value equals {@link javax.swing.ListSelectionModel#SINGLE_INTERVAL_SELECTION}.
     */
    public static final int SINGLE_INTERVAL_SELECTION = 1;
    
    /**
     * A selection mode where any element may be selected and elements added
     * adjacent to selected elements are selected. For convenience, this value
     * equals {@link javax.swing.ListSelectionModel#MULTIPLE_INTERVAL_SELECTION}.
     */
    public static final int MULTIPLE_INTERVAL_SELECTION = 2;
    
    /**
     * A selection mode where any element may be selected and freshly added elements
     * are always deselected. No equivalent policy exists in
     * {@link javax.swing.ListSelectionModel}
     */
    public static final int MULTIPLE_INTERVAL_SELECTION_DEFENSIVE = 103;
    
    /**
     * the source list 
     */
    private final EventList source;
    
    /**
     * the selected view 
     */
    private final SelectedList selectedList;
    
    /**
     * the deselected view 
     */
    private final DeselectedList deselectedList;
    
    /**
     * observe the source list 
     */
    private final SourceListener sourceListener = new SourceListener();
    
    /**
     * the selection state 
     */
    private Barcode barcode = new Barcode();
    
    /**
     * the lead selection index 
     */
    private int leadSelectionIndex = -1;
    
    /**
     * the anchor selection index 
     */
    private int anchorSelectionIndex = -1;
    
    /**
     * the selection mode defines characteristics of the selection 
     */
    private int selectionMode = MULTIPLE_INTERVAL_SELECTION_DEFENSIVE;
    
    /**
     * the current selection colour 
     */
    private Object selected = Barcode.BLACK;
    
    /**
     * the current deselection colour 
     */
    private Object deselected = Barcode.WHITE;
    
    /**
     * the registered SelectionListeners 
     */
    private List selectionListeners = new ArrayList(1);
    
    /**
     * Creates a new ListSelection that listens to changes on the given source.
     * When using this constructor, all elements are deselected by default.
     */
    public ListSelection(EventList source) {
        super();
        this.source = source;
        barcode.add(0, deselected, source.size());
        source.addListEventListener(sourceListener);
        deselectedList = new DeselectedList(source);
        selectedList = new SelectedList(source);
    }
    
    /**
     * Creates a new ListSelection that listens to changes on the given source
     * and initializes selection with the given array of indices.
     */
    public ListSelection(EventList source, int[] initialSelection) {
        this(source);
        select(initialSelection);
    }
    
    /**
     * Handle changes to the source list by adjusting our selection state and
     * the contents of the selected and deselected lists.
     */
    private class SourceListener implements ListEventListener {
        
        private SourceListener() {
            super();
        }
        
        /**
         * {@inheritDoc} 
         */
        public void listChanged(ListEvent listChanges) {
            int minSelectionIndexBefore = getMinSelectionIndex();
            int maxSelectionIndexBefore = getMaxSelectionIndex();
            if (listChanges.isReordering()) {
                selectedList.updates().beginEvent();
                int[] sourceReorderMap = listChanges.getReorderMap();
                int[] selectReorderMap = new int[barcode.colourSize(selected)];
                int[] deselectReorderMap = new int[barcode.colourSize(deselected)];
                Barcode previousBarcode = barcode;
                barcode = new Barcode();
                for (int c = 0; c < sourceReorderMap.length; c++) {
                    Object flag = previousBarcode.get(sourceReorderMap[c]);
                    boolean wasSelected = (flag != deselected);
                    barcode.add(c, flag, 1);
                    if (wasSelected) {
                        int previousIndex = previousBarcode.getColourIndex(sourceReorderMap[c], selected);
                        int currentIndex = barcode.getColourIndex(c, selected);
                        selectReorderMap[currentIndex] = previousIndex;
                    } else {
                        int previousIndex = previousBarcode.getColourIndex(sourceReorderMap[c], deselected);
                        int currentIndex = barcode.getColourIndex(c, deselected);
                        deselectReorderMap[currentIndex] = previousIndex;
                    }
                }
                anchorSelectionIndex = -1;
                leadSelectionIndex = -1;
                selectedList.updates().reorder(selectReorderMap);
                selectedList.updates().commitEvent();
                deselectedList.updates().beginEvent();
                deselectedList.updates().reorder(deselectReorderMap);
                deselectedList.updates().commitEvent();
            } else {
                List savedChanges = new ArrayList();
                selectedList.updates().beginEvent();
                while (listChanges.next()) {
                    int index = listChanges.getIndex();
                    int changeType = listChanges.getType();
                    int previousSelectionIndex = barcode.getColourIndex(index, selected);
                    boolean previouslySelected = previousSelectionIndex != -1;
                    if (changeType == ListEvent.DELETE) {
                        if (previouslySelected) {
                            barcode.remove(index, 1);
                            selectedList.updates().addDelete(previousSelectionIndex);
                        } else {
                            int deselectedIndex = barcode.getColourIndex(index, deselected);
                            savedChanges.add(new DeselectedChange(ListEvent.DELETE, deselectedIndex));
                            barcode.remove(index, 1);
                        }
                    } else if (changeType == ListEvent.INSERT) {
                        if (previouslySelected) {
                            if (selectionMode == SINGLE_INTERVAL_SELECTION || selectionMode == MULTIPLE_INTERVAL_SELECTION) {
                                barcode.add(index, selected, 1);
                                selectedList.updates().addInsert(previousSelectionIndex);
                            } else {
                                barcode.add(index, deselected, 1);
                                int deselectedIndex = barcode.getColourIndex(index, deselected);
                                savedChanges.add(new DeselectedChange(ListEvent.INSERT, deselectedIndex));
                            }
                        } else {
                            barcode.add(index, deselected, 1);
                            int deselectedIndex = barcode.getColourIndex(index, deselected);
                            savedChanges.add(new DeselectedChange(ListEvent.INSERT, deselectedIndex));
                        }
                    } else if (changeType == ListEvent.UPDATE) {
                        if (previouslySelected) {
                            selectedList.updates().addUpdate(previousSelectionIndex);
                        } else {
                            int deselectedIndex = barcode.getColourIndex(index, deselected);
                            savedChanges.add(new DeselectedChange(ListEvent.UPDATE, deselectedIndex));
                        }
                    }
                    anchorSelectionIndex = adjustIndex(anchorSelectionIndex, changeType, index);
                    leadSelectionIndex = adjustIndex(leadSelectionIndex, changeType, index);
                }
                selectedList.updates().commitEvent();
                deselectedList.updates().beginEvent();
                for (int i = 0; i < savedChanges.size(); i++) {
                    DeselectedChange change = (ListSelection.DeselectedChange)savedChanges.get(i);
                    change.fireChange();
                }
                savedChanges.clear();
                deselectedList.updates().commitEvent();
            }
            if (minSelectionIndexBefore != -1 && maxSelectionIndexBefore != -1) {
                int minSelectionIndexAfter = getMinSelectionIndex();
                int maxSelectionIndexAfter = getMaxSelectionIndex();
                int changeStart = minSelectionIndexBefore;
                int changeFinish = maxSelectionIndexBefore;
                if (minSelectionIndexAfter != -1 && minSelectionIndexAfter < changeStart) changeStart = minSelectionIndexAfter;
                if (maxSelectionIndexAfter != -1 && maxSelectionIndexAfter > changeFinish) changeFinish = maxSelectionIndexAfter;
                fireSelectionChanged(changeStart, changeFinish);
            }
        }
    }
    
    /**
     * Adjusts the specified index to the specified change. This is used to adjust
     * the anchor and lead selection indices when list changes occur.
     */
    private int adjustIndex(int indexBefore, int changeType, int changeIndex) {
        if (indexBefore == -1) return -1;
        if (changeType == ListEvent.DELETE) {
            if (changeIndex < indexBefore) return indexBefore - 1; else if (changeIndex == indexBefore) return -1; else return indexBefore;
        } else if (changeType == ListEvent.UPDATE) {
            return indexBefore;
        } else if (changeType == ListEvent.INSERT) {
            if (changeIndex <= indexBefore) return indexBefore + 1; else return indexBefore;
        } else {
            throw new IllegalStateException();
        }
    }
    
    /**
     * Provides access to an {@link EventList} that contains only selected values.
     */
    public EventList getSelected() {
        return selectedList;
    }
    
    /**
     * Provides access to an {@link EventList} that contains only deselected values.
     */
    public EventList getDeselected() {
        return deselectedList;
    }
    
    /**
     * Get the {@link EventList} that selection is being managed for.
     */
    public EventList getSource() {
        return source;
    }
    
    /**
     * Inverts the current selection.
     */
    public void invertSelection() {
        if (selected == Barcode.BLACK) {
            selected = Barcode.WHITE;
            deselected = Barcode.BLACK;
        } else {
            selected = Barcode.BLACK;
            deselected = Barcode.WHITE;
        }
        anchorSelectionIndex = -1;
        leadSelectionIndex = -1;
        selectedList.updates().beginEvent();
        selectedList.updates().addDelete(0, barcode.colourSize(deselected) - 1);
        selectedList.updates().addInsert(0, barcode.colourSize(selected) - 1);
        selectedList.updates().commitEvent();
        deselectedList.updates().beginEvent();
        deselectedList.updates().addDelete(0, barcode.colourSize(selected) - 1);
        deselectedList.updates().addInsert(0, barcode.colourSize(deselected) - 1);
        deselectedList.updates().commitEvent();
        fireSelectionChanged(0, source.size() - 1);
    }
    
    /**
     * Returns whether or not the item with the given source index
     * is selected.
     */
    public boolean isSelected(int sourceIndex) {
        if (sourceIndex < 0 || sourceIndex >= source.size()) {
            return false;
        }
        return barcode.getColourIndex(sourceIndex, selected) != -1;
    }
    
    /**
     * Deselects the element at the given index.
     */
    public void deselect(int index) {
        deselect(index, index);
    }
    
    /**
     * Deselects all of the elements within the given range.
     */
    public void deselect(int start, int end) {
        if (start == -1 || end == -1) {
            return;
        } else if (selectionMode == SINGLE_SELECTION) {
            int selectedIndex = getMaxSelectionIndex();
            if (selectedIndex >= start && selectedIndex <= end) {
                deselectAll();
            }
            return;
        } else if (selectionMode == SINGLE_INTERVAL_SELECTION && start > getMinSelectionIndex()) {
            end = Math.max(end, getMaxSelectionIndex());
        }
        anchorSelectionIndex = start;
        leadSelectionIndex = end;
        setSubRangeOfRange(false, start, end, -1, -1);
    }
    
    /**
     * Deselects all of the elements in the given array of indices.  The
     * array must contain indices in sorted, ascending order.
     */
    public void deselect(int[] indices) {
        List deselections = new ArrayList();
        int firstAffectedIndex = -1;
        int lastAffectedIndex = -1;
        selectedList.updates().beginEvent();
        int currentIndex = 0;
        for (BarcodeIterator i = barcode.iterator(); i.hasNext() && currentIndex != indices.length; ) {
            Object value = i.next();
            if (i.getIndex() == indices[currentIndex]) {
                if (value == selected) {
                    if (firstAffectedIndex == -1) firstAffectedIndex = i.getIndex();
                    lastAffectedIndex = i.getIndex();
                    int selectedIndex = i.getColourIndex(selected);
                    deselections.add(new DeselectedChange(ListEvent.INSERT, i.set(deselected)));
                    selectedList.updates().addDelete(selectedIndex);
                }
                currentIndex++;
            }
        }
        selectedList.updates().commitEvent();
        deselectedList.updates().beginEvent();
        for (int i = 0; i < deselections.size(); i++) {
            DeselectedChange change = (ListSelection.DeselectedChange)deselections.get(i);
            change.fireChange();
        }
        deselectedList.updates().commitEvent();
        if (firstAffectedIndex > -1) fireSelectionChanged(firstAffectedIndex, lastAffectedIndex);
    }
    
    /**
     * Deselects all elements.
     */
    public void deselectAll() {
        int selectionChangeSize = barcode.colourSize(selected);
        if (selectionChangeSize == 0) return;
        int firstAffectedIndex = -1;
        int lastAffectedIndex = -1;
        deselectedList.updates().beginEvent();
        for (BarcodeIterator i = barcode.iterator(); i.hasNextColour(selected); ) {
            i.nextColour(selected);
            if (firstAffectedIndex == -1) firstAffectedIndex = i.getIndex();
            lastAffectedIndex = i.getIndex();
            deselectedList.updates().addInsert(i.getIndex());
        }
        barcode.clear();
        barcode.add(0, deselected, source.size());
        deselectedList.updates().commitEvent();
        selectedList.updates().beginEvent();
        selectedList.updates().addDelete(0, selectionChangeSize - 1);
        selectedList.updates().commitEvent();
        fireSelectionChanged(firstAffectedIndex, lastAffectedIndex);
    }
    
    /**
     * Selects the element at the given index.
     */
    public void select(int index) {
        select(index, index);
    }
    
    /**
     * Selects all of the elements within the given range.
     */
    public void select(int start, int end) {
        if (start == -1 || end == -1) {
            return;
        } else if (selectionMode == SINGLE_SELECTION) {
            setSelection(start);
            return;
        } else if (selectionMode == SINGLE_INTERVAL_SELECTION) {
            boolean overlap = false;
            int minSelectedIndex = getMinSelectionIndex();
            int maxSelectedIndex = getMaxSelectionIndex();
            if (minSelectedIndex - 1 <= start && start <= maxSelectedIndex + 1) overlap = true;
            if (minSelectedIndex - 1 <= end && end <= maxSelectedIndex + 1) overlap = true;
            if (!overlap) {
                setSelection(start, end);
                return;
            }
        }
        anchorSelectionIndex = start;
        leadSelectionIndex = end;
        setSubRangeOfRange(true, start, end, -1, -1);
    }
    
    /**
     * Selects all of the elements in the given array of indices.  The
     * array must contain indices in sorted, ascending order.
     */
    public void select(int[] indices) {
        List selections = new ArrayList();
        int firstAffectedIndex = -1;
        int lastAffectedIndex = -1;
        selectedList.updates().beginEvent();
        int currentIndex = 0;
        for (BarcodeIterator i = barcode.iterator(); i.hasNext() && currentIndex != indices.length; ) {
            Object value = i.next();
            if (i.getIndex() == indices[currentIndex]) {
                if (value != selected) {
                    if (firstAffectedIndex == -1) firstAffectedIndex = i.getIndex();
                    lastAffectedIndex = i.getIndex();
                    int deselectedIndex = i.getColourIndex(deselected);
                    selections.add(new DeselectedChange(ListEvent.DELETE, deselectedIndex));
                    selectedList.updates().addInsert(i.set(selected));
                }
                currentIndex++;
            }
        }
        selectedList.updates().commitEvent();
        deselectedList.updates().beginEvent();
        for (int i = 0; i < selections.size(); i++) {
            DeselectedChange change = (ListSelection.DeselectedChange)selections.get(i);
            change.fireChange();
        }
        deselectedList.updates().commitEvent();
        if (firstAffectedIndex > -1) fireSelectionChanged(firstAffectedIndex, lastAffectedIndex);
    }
    
    /**
     * Select the specified element, if it exists.
     *
     * @return the index of the newly selected element, or -1 if no
     *     element was found.
     */
    public int select(Object value) {
        int index = source.indexOf(value);
        if (index != -1) select(index);
        return index;
    }
    
    /**
     * Select all of the specified values.
     *
     * @return <code>true</code> if the selection changed as a result of the call.
     */
    public boolean select(Collection values) {
        SortedSet indicesToSelect = new TreeSet();
        for (Iterator v = values.iterator(); v.hasNext(); ) {
            Object value = v.next();
            int index = source.indexOf(value);
            if (index == -1) continue;
            indicesToSelect.add(new Integer(index));
        }
        if (indicesToSelect.isEmpty()) return false;
        int[] indicesToSelectAsInts = new int[indicesToSelect.size()];
        int arrayIndex = 0;
        for (Iterator i = indicesToSelect.iterator(); i.hasNext(); ) {
            Integer selectIndex = (Integer)i.next();
            indicesToSelectAsInts[arrayIndex] = selectIndex.intValue();
            arrayIndex++;
        }
        int selectionSizeBefore = getSelected().size();
        select(indicesToSelectAsInts);
        int selectionSizeAfter = getSelected().size();
        return selectionSizeAfter > selectionSizeBefore;
    }
    
    /**
     * Selects all elements.
     */
    public void selectAll() {
        int deselectionChangeSize = barcode.colourSize(deselected);
        if (deselectionChangeSize == 0) return;
        int firstAffectedIndex = -1;
        int lastAffectedIndex = -1;
        selectedList.updates().beginEvent();
        for (BarcodeIterator i = barcode.iterator(); i.hasNextColour(deselected); ) {
            i.nextColour(deselected);
            if (firstAffectedIndex == -1) firstAffectedIndex = i.getIndex();
            lastAffectedIndex = i.getIndex();
            selectedList.updates().addInsert(i.getIndex());
        }
        barcode.clear();
        barcode.add(0, selected, source.size());
        selectedList.updates().commitEvent();
        deselectedList.updates().beginEvent();
        deselectedList.updates().addDelete(0, deselectionChangeSize - 1);
        deselectedList.updates().commitEvent();
        fireSelectionChanged(firstAffectedIndex, lastAffectedIndex);
    }
    
    /**
     * Sets the selection to be only the element at the given index.  If the
     * given index is -1, the selection will be cleared.
     */
    public void setSelection(int index) {
        setSelection(index, index);
    }
    
    /**
     * Sets the selection to be only elements within the given range.  If the
     * endpoints of the range are -1, the selection will be cleared.
     */
    public void setSelection(int start, int end) {
        if (start == -1 || end == -1) {
            deselectAll();
            return;
        } else if (selectionMode == SINGLE_SELECTION) {
            end = start;
        }
        anchorSelectionIndex = start;
        leadSelectionIndex = end;
        setSubRangeOfRange(true, start, end, getMinSelectionIndex(), getMaxSelectionIndex());
    }
    
    /**
     * Sets the selection to be only the element in the given array of indices.
     * Unlike {@link #setSelection(int)} and {@link #setSelection(int,int)},
     * providing a value of -1 is an error.  The array must contain indices in
     * sorted, ascending order.
     */
    public void setSelection(int[] indices) {
        if (indices.length == 0) {
            deselectAll();
            return;
        }
        List changes = new ArrayList();
        int firstAffectedIndex = -1;
        int lastAffectedIndex = -1;
        selectedList.updates().beginEvent();
        int currentIndex = 0;
        for (BarcodeIterator i = barcode.iterator(); i.hasNext(); ) {
            Object value = i.next();
            if (i.getIndex() == indices[currentIndex]) {
                if (value != selected) {
                    if (firstAffectedIndex == -1) firstAffectedIndex = i.getIndex();
                    lastAffectedIndex = i.getIndex();
                    int deselectedIndex = i.getColourIndex(deselected);
                    changes.add(new DeselectedChange(ListEvent.DELETE, deselectedIndex));
                    selectedList.updates().addInsert(i.set(selected));
                }
                if (currentIndex < indices.length - 1) currentIndex++;
            } else if (value == selected) {
                if (firstAffectedIndex == -1) firstAffectedIndex = i.getIndex();
                lastAffectedIndex = i.getIndex();
                int selectedIndex = i.getColourIndex(selected);
                selectedList.updates().addDelete(selectedIndex);
                changes.add(new DeselectedChange(ListEvent.INSERT, i.set(deselected)));
            }
        }
        selectedList.updates().commitEvent();
        deselectedList.updates().beginEvent();
        for (int i = 0; i < changes.size(); i++) {
            DeselectedChange change = (ListSelection.DeselectedChange)changes.get(i);
            change.fireChange();
        }
        deselectedList.updates().commitEvent();
        if (firstAffectedIndex > -1) fireSelectionChanged(firstAffectedIndex, lastAffectedIndex);
    }
    
    /**
     * Return the anchor of the current selection.
     */
    public int getAnchorSelectionIndex() {
        return anchorSelectionIndex;
    }
    
    /**
     * Set the anchor selection index.
     */
    public void setAnchorSelectionIndex(int anchorSelectionIndex) {
        this.anchorSelectionIndex = anchorSelectionIndex;
        if (anchorSelectionIndex == -1 || leadSelectionIndex == -1) {
            deselectAll();
        } else if (selectionMode == SINGLE_SELECTION) {
            setSubRangeOfRange(true, anchorSelectionIndex, anchorSelectionIndex, getMinSelectionIndex(), getMaxSelectionIndex());
        } else if (selectionMode == SINGLE_INTERVAL_SELECTION) {
            setSubRangeOfRange(true, anchorSelectionIndex, leadSelectionIndex, getMinSelectionIndex(), getMaxSelectionIndex());
        } else {
            setSubRangeOfRange(true, anchorSelectionIndex, leadSelectionIndex, -1, -1);
        }
    }
    
    /**
     * Return the lead of the current selection.
     */
    public int getLeadSelectionIndex() {
        return leadSelectionIndex;
    }
    
    /**
     * Set the lead selection index.
     */
    public void setLeadSelectionIndex(int leadSelectionIndex) {
        int originalLeadIndex = this.leadSelectionIndex;
        this.leadSelectionIndex = leadSelectionIndex;
        if (leadSelectionIndex == -1 || anchorSelectionIndex == -1) {
            deselectAll();
        } else if (selectionMode == SINGLE_SELECTION) {
            setSubRangeOfRange(true, leadSelectionIndex, leadSelectionIndex, getMinSelectionIndex(), getMaxSelectionIndex());
        } else if (selectionMode == SINGLE_INTERVAL_SELECTION) {
            setSubRangeOfRange(true, anchorSelectionIndex, leadSelectionIndex, getMinSelectionIndex(), getMaxSelectionIndex());
        } else {
            setSubRangeOfRange(true, anchorSelectionIndex, leadSelectionIndex, anchorSelectionIndex, originalLeadIndex);
        }
    }
    
    /**
     * Set the selection mode.
     */
    public void setSelectionMode(int selectionMode) {
        this.selectionMode = selectionMode;
        setSelection(getMinSelectionIndex(), getMaxSelectionIndex());
    }
    
    /**
     * Returns the current selection mode.
     */
    public int getSelectionMode() {
        return selectionMode;
    }
    
    /**
     * Returns the first selected index or -1 if nothing is selected.
     */
    public int getMinSelectionIndex() {
        if (barcode.colourSize(selected) == 0) return -1;
        return barcode.getIndex(0, selected);
    }
    
    /**
     * Returns the last selected index or -1 if nothing is selected.
     */
    public int getMaxSelectionIndex() {
        if (barcode.colourSize(selected) == 0) return -1;
        return barcode.getIndex(barcode.colourSize(selected) - 1, selected);
    }
    
    /**
     * Walks through the union of the specified ranges. All values in the
     * first range are given the specified selection value. All values in
     * the second range but not in the first range are given the opposite
     * selection value. In effect, the first range and the intersection of
     * the two ranges are given the specified value. The parts that are
     * in the second range but not in the first range (ie. everything else)
     * is given the opposite selection value.
     *
     * @param select true to set values in the first range as selected and
     *      the other values in the second range as not selected. false to
     *      set values in the first range as not selected and the other
     *      values in the second range as selected.
     * @param changeIndex0 one end of the first range, inclusive.
     * @param changeIndex1 the opposite end of the first range, inclusive.
     *      This may be lower than changeIndex0.
     * @param invertIndex0 one end of the second range, inclusive. To
     *      specify an empty second range, specify -1 for this value.
     * @param invertIndex1 the opposite end of the second range, inclusive.
     *      This may be lower than invertIndex0. To specify an empty second
     *      range, specify -1 for this value.
     */
    private void setSubRangeOfRange(boolean select, int changeIndex0, int changeIndex1, int invertIndex0, int invertIndex1) {
        if (changeIndex0 >= source.size() || changeIndex1 >= source.size() || ((changeIndex0 == -1 || changeIndex1 == -1) && changeIndex0 != changeIndex1)) {
            throw new IndexOutOfBoundsException("Invalid range for selection: " + changeIndex0 + "-" + changeIndex1 + ", list size is " + source.size());
        }
        if (invertIndex0 >= source.size() || invertIndex1 >= source.size() || ((invertIndex0 == -1 || invertIndex1 == -1) && invertIndex0 != invertIndex1)) {
            throw new IndexOutOfBoundsException("Invalid range for invert selection: " + invertIndex0 + "-" + invertIndex1 + ", list size is " + source.size());
        }
        if (changeIndex0 == -1 && changeIndex1 == -1) {
            if (invertIndex0 == -1 && invertIndex1 == -1) return;
            changeIndex0 = invertIndex0;
            changeIndex1 = invertIndex1;
            select = !select;
        }
        if (invertIndex0 == -1 && invertIndex1 == -1) {
            invertIndex0 = changeIndex0;
            invertIndex1 = changeIndex1;
        }
        int minChangeIndex = Math.min(changeIndex0, changeIndex1);
        int maxChangeIndex = Math.max(changeIndex0, changeIndex1);
        int minInvertIndex = Math.min(invertIndex0, invertIndex1);
        int maxInvertIndex = Math.max(invertIndex0, invertIndex1);
        int minUnionIndex = Math.min(minChangeIndex, minInvertIndex);
        int maxUnionIndex = Math.max(maxChangeIndex, maxInvertIndex);
        int minChangedIndex = maxUnionIndex + 1;
        int maxChangedIndex = minUnionIndex - 1;
        selectedList.updates().beginEvent();
        List savedChanges = new ArrayList();
        for (int i = minUnionIndex; i <= maxUnionIndex; i++) {
            int selectionIndex = barcode.getColourIndex(i, selected);
            boolean selectedBefore = (selectionIndex != -1);
            boolean inChangeRange = (i >= minChangeIndex && i <= maxChangeIndex);
            boolean selectedAfter = (inChangeRange == select);
            if (selectedBefore != selectedAfter) {
                if (i < minChangedIndex) minChangedIndex = i;
                if (i > maxChangedIndex) maxChangedIndex = i;
                if (selectedBefore) {
                    barcode.set(i, deselected, 1);
                    selectedList.updates().addDelete(selectionIndex);
                    savedChanges.add(new DeselectedChange(ListEvent.INSERT, i - selectionIndex));
                } else {
                    barcode.set(i, selected, 1);
                    int newSelectionIndex = barcode.getColourIndex(i, selected);
                    selectedList.updates().addInsert(newSelectionIndex);
                    savedChanges.add(new DeselectedChange(ListEvent.DELETE, i - newSelectionIndex));
                }
            }
        }
        selectedList.updates().commitEvent();
        deselectedList.updates().beginEvent();
        for (int i = 0; i < savedChanges.size(); i++) {
            DeselectedChange change = (ListSelection.DeselectedChange)savedChanges.get(i);
            change.fireChange();
        }
        savedChanges.clear();
        deselectedList.updates().commitEvent();
        if (minChangedIndex <= maxChangedIndex) fireSelectionChanged(minChangedIndex, maxChangedIndex);
    }
    
    /**
     * Register a {@link ca.odell.glazedlists.ListSelection.Listener Listener}
     * that will be notified when selection is changed.
     */
    public void addSelectionListener(Listener selectionListener) {
        selectionListeners.add(selectionListener);
    }
    
    /**
     * Remove a {@link ca.odell.glazedlists.ListSelection.Listener Listener}
     * so that it will no longer be notified when selection changes.
     */
    public void removeSelectionListener(Listener selectionListener) {
        selectionListeners.remove(selectionListener);
    }
    
    /**
     * Fire changes in selection to all registered listeners.
     */
    private void fireSelectionChanged(int start, int end) {
        for (Iterator i = selectionListeners.iterator(); i.hasNext(); ) {
            Listener listener = (ListSelection.Listener)i.next();
            listener.selectionChanged(start, end);
        }
    }
    
    /**
     * Disposes of this ListSelection freeing up it's resources for
     * garbage collection.  It is an error to use a ListSelection after
     * dispose() has been called.
     */
    public void dispose() {
        source.removeListEventListener(sourceListener);
        selectionListeners.clear();
    }
    
    /**
     * A generic interface to respond to changes in selection that doesn't
     * require including a particular GUI toolkit.
     */
    public interface Listener {
        
        /**
         * Notifies this SelectionListener of a change in selection.
         *
         * @param changeStart The first zero-relative index affected by a change in selection.
         * @param changeEnd   The last zero-relative index affected by a change in selection.
         */
        public void selectionChanged(int changeStart, int changeEnd);
    }
    
    /**
     * This is a bit of a hack, but it gets around the cannot start a new Event
     * while an Event is in progress expection from the ListEventAssembler
     */
    private final class DeselectedChange {
        
        /**
         * the type of event corresponding to the types defined in ListEvent
         */
        private int type = 0;
        
        /**
         * the index the change occurred at 
         */
        private int index = 0;
        
        /**
         * Creates a new DeselectedChange which bears a striking resemblance to
         * ListEvent.
         */
        public DeselectedChange(int type, int index) {
            super();
            this.type = type;
            this.index = index;
        }
        
        /**
         * Fires the change event on the Deselected list.
         */
        public void fireChange() {
            deselectedList.updates().addChange(type, index);
        }
    }
    
    /**
     * The {@link EventList} that contains only values that are not currently
     * selected.
     */
    private final class SelectedList extends TransformedList {
        
        /**
         * Creates an {@link EventList} that provides a view of the
         * deselected items in a ListSelection.
         */
        SelectedList(EventList source) {
            super(source);
        }
        
        /**
         * {@inheritDoc} 
         */
        public int size() {
            return barcode.colourSize(selected);
        }
        
        /**
         * {@inheritDoc} 
         */
        protected int getSourceIndex(int mutationIndex) {
            return barcode.getIndex(mutationIndex, selected);
        }
        
        /**
         * {@inheritDoc} 
         */
        public void listChanged(ListEvent listChanges) {
        }
        
        /**
         * This allows access to the EventAssembler for this list.
         */
        public ListEventAssembler updates() {
            return updates;
        }
        
        /**
         * {@inheritDoc} 
         */
        protected boolean isWritable() {
            return true;
        }
        
        /**
         * A no-op dispose method to prevent the user from shooting themselves
         * in the foot. To dispose a {@link ListSelection}, call
         * {@link ListSelection#dispose()} on that class directly.
         */
        public void dispose() {
        }
    }
    
    /**
     * The {@link EventList} that contains only values that are not currently
     * selected.
     */
    private final class DeselectedList extends TransformedList {
        
        /**
         * Creates an {@link EventList} that provides a view of the
         * deselected items in a ListSelection.
         */
        DeselectedList(EventList source) {
            super(source);
        }
        
        /**
         * {@inheritDoc} 
         */
        public int size() {
            return barcode.colourSize(deselected);
        }
        
        /**
         * {@inheritDoc} 
         */
        protected int getSourceIndex(int mutationIndex) {
            return barcode.getIndex(mutationIndex, deselected);
        }
        
        /**
         * {@inheritDoc} 
         */
        public void listChanged(ListEvent listChanges) {
        }
        
        /**
         * This allows access to the EventAssembler for this list.
         */
        public ListEventAssembler updates() {
            return updates;
        }
        
        /**
         * {@inheritDoc} 
         */
        protected boolean isWritable() {
            return true;
        }
        
        /**
         * A no-op dispose method to prevent the user from shooting themselves
         * in the foot. To dispose a {@link ListSelection}, call
         * {@link ListSelection#dispose()} on that class directly.
         */
        public void dispose() {
        }
    }
}
