package ca.odell.glazedlists.impl.adt;

import java.util.*;

/**
 * A tree node that can be accessed either in sorted order or by
 * index.
 *
 * <p>This tree-node uses AVL-Trees to ensure that access is always
 * logarithmic in terms of the size of the tree. AVL Trees use
 * rotations (single and double) when the height of a pair of
 * subtrees do not match in order to guarantee a bound on the
 * difference in their height. This bound can be shown to provide
 * an overall bound on the access time on the tree.
 *
 * <p>As of October 9, 2005, this class can be used to create
 * <i>partially sorted</i> trees. This means that some subset of the nodes
 * are not considered when sorting. The purpose of this is to provide a
 * reasonable data store when new elements should be added in sorted order,
 * but old elements are not to be moved once added. To mark a node as unsorted,
 * use the
 *
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public final class IndexedTreeNode {
    
    /**
     * a token node 
     */
    public static final IndexedTreeNode EMPTY_NODE = new IndexedTreeNode(null, "EMPTY_NODE");
    
    /**
     * the parent node, used to delete from leaf up 
     */
    IndexedTreeNode parent;
    
    /**
     * the left and right child nodes 
     */
    IndexedTreeNode left = null;
    IndexedTreeNode right = null;
    
    /**
     * the size of the left and right subtrees 
     */
    int leftSize = 0;
    int rightSize = 0;
    
    /**
     * the height of this subtree 
     */
    private int height = 0;
    
    /**
     * The value of this node, which should be greater than all values in
     * the left subtree but less than all values in the right subtree.
     */
    private Object value;
    
    /**
     * Whether to use the node's value when doing comparisons. Otherwise
     * the node value is a placeholder for a value and should <strong>not</strong>
     * be used when doing comparisons on operations like
     * {@link #getShallowestNodeWithValue} or {@link #insert}.
     */
    private boolean sorted = true;
    
    /**
     * Creates a new IndexedTreeNode with the specified parent node.
     */
    IndexedTreeNode(IndexedTreeNode parent, Object value) {
        super();
        this.parent = parent;
        this.value = value;
    }
    
    /**
     * Gets the value of this tree node.
     */
    public Object getValue() {
        return value;
    }
    
    /**
     * Sets the value of this tree node. <strong>Warning:</strong> changing
     * the value of a node in a sorted tree may cause sorting to break
     * miserably.
     */
    public void setValue(Object value) {
        this.value = value;
    }
    
    /**
     * Set this node to be used or ignored when sorting other values in the
     * tree. Unsorted nodes are useful when it is necessary to keep elements
     * in place even after their sort-order may have changed.
     */
    public void setSorted(boolean sorted) {
        this.sorted = sorted;
    }
    
    /**
     * Whether this node is sorted.
     */
    public boolean isSorted() {
        return sorted;
    }
    
    /**
     * Gets the object with the specified index in the tree.
     */
    IndexedTreeNode getNodeWithIndex(int index) {
        if (index < leftSize) {
            return left.getNodeWithIndex(index);
        } else if (index > leftSize) {
            return right.getNodeWithIndex(index - (leftSize + 1));
        } else {
            return this;
        }
    }
    
    /**
     * @return <code>true</code> if this is on the left side of its parent.
     */
    private boolean isRightChild() {
        return parent != null && parent.right == this;
    }
    
    /**
     * @return <code>true</code> if this is on the right side of its parent.
     */
    private boolean isLeftChild() {
        return parent != null && parent.left == this;
    }
    
    /**
     * Whether the search value sorts on the left, right or center.
     *
     * @return 0 if this element equals the search value, a negative number if
     *      the search value belongs on the left and a positive number if the
     *      search value belongs on the right.
     */
    private int getSortSide(Comparator comparator, Object searchValue) {
        IndexedTreeNode currentFollower = this;
        while (true) {
            if (currentFollower == null) {
                return -1;
            } else if (currentFollower.sorted) {
                return comparator.compare(searchValue, currentFollower.value);
            }
            currentFollower = currentFollower.next();
        }
    }
    
    /**
     * @return the node whose index is one greater than this node, or <code>null</code>
     *      if this is the last node in the tree.
     */
    public IndexedTreeNode next() {
        IndexedTreeNode result;
        if (right != null) {
            result = right;
            while (result.left != null) {
                result = result.left;
            }
        } else if (parent == null) {
            result = null;
        } else if (isLeftChild()) {
            result = parent;
        } else if (isRightChild()) {
            IndexedTreeNode parentOfRightChild = this.parent;
            while (parentOfRightChild.isRightChild()) {
                parentOfRightChild = parentOfRightChild.parent;
            }
            result = parentOfRightChild.parent;
        } else {
            throw new IllegalStateException();
        }
        return result;
    }
    
    /**
     * @return the node whose index is one less than this node, or <code>null</code>
     *      if this is the first node in the tree.
     */
    public IndexedTreeNode previous() {
        IndexedTreeNode result;
        if (left != null) {
            result = left;
            while (result.right != null) {
                result = result.right;
            }
        } else if (parent == null) {
            result = null;
        } else if (isRightChild()) {
            result = parent;
        } else if (isLeftChild()) {
            IndexedTreeNode parentOfLeftChild = this.parent;
            while (parentOfLeftChild.isLeftChild()) {
                parentOfLeftChild = parentOfLeftChild.parent;
            }
            result = parentOfLeftChild.parent;
        } else {
            throw new IllegalStateException();
        }
        return result;
    }
    
    /**
     * Return the first node found such that {@link Comparator#compare} ranks
     * the specified object equal to that node's value. It is up to the caller
     * to narrow in on a more specific node to be used by methods like
     * {@link List#indexOf} etc.
     */
    IndexedTreeNode getShallowestNodeWithValue(Comparator comparator, Object object, boolean approximate) {
        int sortSide = getSortSide(comparator, object);
        if (sortSide == 0) {
            return this;
        }
        IndexedTreeNode recurseNode = (sortSide < 0) ? left : right;
        if (recurseNode != null) {
            return recurseNode.getShallowestNodeWithValue(comparator, object, approximate);
        }
        return approximate ? this : null;
    }
    
    /**
     * Retrieves the size of this subtree.
     */
    int size() {
        return leftSize + 1 + rightSize;
    }
    
    /**
     * Retrieves the height of this subtree.
     */
    int height() {
        return height;
    }
    
    /**
     * Retrieves the subtree node with the largest value.
     */
    IndexedTreeNode getLargestChildNode() {
        if (rightSize > 0) return right.getLargestChildNode(); else return this;
    }
    
    /**
     * Retrieves the subtree node with the smallest value.
     */
    IndexedTreeNode getSmallestChildNode() {
        if (leftSize > 0) return left.getSmallestChildNode(); else return this;
    }
    
    /**
     * Gets the index of the current node, based on a recurrsive
     * path up the tree.
     */
    public int getIndex() {
        return getIndex(null);
    }
    
    private int getIndex(IndexedTreeNode child) {
        if (child == left) {
            if (parent != null) return parent.getIndex(this);
            return 0;
        } else if (child == null) {
            if (parent != null) return parent.getIndex(this) + leftSize;
            return leftSize;
        } else if (child == right) {
            if (parent != null) return parent.getIndex(this) + leftSize + 1;
            return leftSize + 1;
        }
        throw new IndexOutOfBoundsException(this + " cannot get the index of a subtree that does not exist on this node!");
    }
    
    /**
     * Inserts the specified object into the tree in sorted order.
     *
     * @return the IndexedTreeNode node where the object was inserted. This
     *      node can be used to call the deleteUp() method, which will
     *      delete the node from it's parent tree. It is also possible to
     *      use the getIndex() method on the node to discover what the sorted
     *      index of the value is.
     */
    IndexedTreeNode insert(IndexedTree host, Object value) {
        boolean insertOnLeft = getSortSide(host.getComparator(), value) < 0;
        IndexedTreeNode subtree = insertOnLeft ? left : right;
        final IndexedTreeNode inserted;
        if (insertOnLeft) this.leftSize++; else this.rightSize++;
        if (subtree == null) {
            inserted = new IndexedTreeNode(this, value);
            if (insertOnLeft) this.left = inserted; else this.right = inserted;
            inserted.ensureAVL(host);
        } else {
            inserted = subtree.insert(host, value);
        }
        return inserted;
    }
    
    /**
     * Inserts the specified object into the tree with the specified index.
     *
     * @return the IndexedTreeNode node where the object was inserted. This
     *      node can be used to call the deleteUp() method, which will
     *      delete the node from it's parent tree. It is also possible to call
     *      the getIndex() method on the node. As new nodes are inserted, the
     *      index will shift. The getIndex() method can be used to get the
     *      current index of the node at any time.
     */
    IndexedTreeNode insert(IndexedTree host, int index, Object value) {
        boolean insertOnLeft = index <= leftSize;
        IndexedTreeNode subtree = insertOnLeft ? left : right;
        final IndexedTreeNode inserted;
        if (insertOnLeft) this.leftSize++; else this.rightSize++;
        if (subtree == null) {
            inserted = new IndexedTreeNode(this, value);
            if (insertOnLeft) this.left = inserted; else this.right = inserted;
            inserted.ensureAVL(host);
        } else {
            int recurseIndex = insertOnLeft ? index : index - leftSize - 1;
            inserted = subtree.insert(host, recurseIndex, value);
        }
        return inserted;
    }
    
    /**
     * Recurses down from the root and removes the node at the given index.
     */
    IndexedTreeNode removeNode(IndexedTree host, int index) {
        if (index < leftSize) {
            leftSize--;
            return left.removeNode(host, index);
        } else if (index > leftSize) {
            rightSize--;
            return right.removeNode(host, index - (leftSize + 1));
        }
        if (leftSize == 0 && rightSize == 0) {
            if (parent != null) {
                parent.replaceChildNode(this, null);
                parent.ensureAVL(host);
            } else {
                host.setRootNode(null);
            }
        } else if (leftSize > 0 && rightSize == 0) {
            left.parent = parent;
            if (parent != null) {
                parent.replaceChildNode(this, left);
                parent.ensureAVL(host);
            } else {
                host.setRootNode(left);
            }
        } else if (leftSize == 0 && rightSize > 0) {
            right.parent = parent;
            if (parent != null) {
                parent.replaceChildNode(this, right);
                parent.ensureAVL(host);
            } else {
                host.setRootNode(right);
            }
        } else {
            IndexedTreeNode replacement = null;
            if (leftSize > rightSize) {
                leftSize--;
                replacement = left.pruneLargestChild(host);
            } else {
                rightSize--;
                replacement = right.pruneSmallestChild(host);
            }
            replacement.left = left;
            replacement.leftSize = leftSize;
            if (left != null) left.parent = replacement;
            replacement.right = right;
            replacement.rightSize = rightSize;
            if (right != null) right.parent = replacement;
            replacement.height = height;
            replacement.parent = parent;
            if (parent != null) {
                parent.replaceChildNode(this, replacement);
            } else {
                host.setRootNode(replacement);
            }
        }
        clearNodeValues();
        return this;
    }
    
    /**
     * Removes the smallest child of the subtree rooted at this and returns it.
     */
    IndexedTreeNode pruneSmallestChild(IndexedTree host) {
        if (leftSize > 0) {
            leftSize--;
            return left.pruneSmallestChild(host);
        }
        IndexedTreeNode replacement = null;
        if (rightSize != 0) {
            replacement = right;
            right.parent = parent;
        }
        if (parent != null) {
            parent.replaceChildNode(this, replacement);
            parent.ensureAVL(host);
        } else {
            host.setRootNode(replacement);
        }
        clearNodeValues();
        return this;
    }
    
    /**
     * Removes the largest child of the subtree rooted at this and returns it.
     */
    IndexedTreeNode pruneLargestChild(IndexedTree host) {
        if (rightSize > 0) {
            rightSize--;
            return right.pruneLargestChild(host);
        }
        IndexedTreeNode replacement = null;
        if (leftSize != 0) {
            replacement = left;
            left.parent = parent;
        }
        if (parent != null) {
            parent.replaceChildNode(this, replacement);
            parent.ensureAVL(host);
        } else {
            host.setRootNode(replacement);
        }
        clearNodeValues();
        return this;
    }
    
    /**
     * Unlinks this node from the sorted tree. This may cause the tree to
     * rotate nodes using AVL rotations.
     */
    public void removeFromTree(IndexedTree host) {
        assert (value != null);
        if (leftSize == 0 && rightSize == 0) {
            if (parent != null) {
                parent.notifyChildNodeRemoved(this);
                parent.replaceChildNode(this, null);
                parent.ensureAVL(host);
            } else {
                host.setRootNode(null);
            }
        } else if (leftSize > 0 && rightSize == 0) {
            left.parent = parent;
            if (parent != null) {
                parent.notifyChildNodeRemoved(this);
                parent.replaceChildNode(this, left);
                parent.ensureAVL(host);
            } else {
                host.setRootNode(left);
            }
        } else if (leftSize == 0 && rightSize > 0) {
            right.parent = parent;
            if (parent != null) {
                parent.notifyChildNodeRemoved(this);
                parent.replaceChildNode(this, right);
                parent.ensureAVL(host);
            } else {
                host.setRootNode(right);
            }
        } else {
            IndexedTreeNode middle = null;
            if (leftSize > rightSize) {
                middle = left.getLargestChildNode();
            } else {
                middle = right.getSmallestChildNode();
            }
            middle.removeFromTree(host);
            assert (middle.leftSize == 0 && middle.rightSize == 0);
            middle.left = left;
            middle.leftSize = leftSize;
            if (left != null) left.parent = middle;
            middle.right = right;
            middle.rightSize = rightSize;
            if (right != null) right.parent = middle;
            middle.height = height;
            middle.parent = parent;
            if (parent != null) {
                parent.replaceChildNode(this, middle);
                parent.ensureAVL(host);
            } else {
                host.setRootNode(middle);
            }
        }
        clearNodeValues();
    }
    
    /**
     * Clears all values related to how this node is linked in the tree, but
     * does NOT clear the value of the node available via setValue().
     */
    private void clearNodeValues() {
        parent = null;
        left = null;
        leftSize = 0;
        right = null;
        rightSize = 0;
    }
    
    /**
     * Notifies that a node has been removed from the specified subtree.
     * This simply decrements the count on that subtree.
     */
    private void notifyChildNodeRemoved(IndexedTreeNode subtree) {
        if (subtree == left) leftSize--; else if (subtree == right) rightSize--; else throw new IllegalArgumentException(this + " cannot remove a subtree that does not exist on this node!");
        if (parent != null) parent.notifyChildNodeRemoved(this);
    }
    
    /**
     * Replaces the specified child with a new child.
     */
    private void replaceChildNode(IndexedTreeNode original, IndexedTreeNode replacement) {
        if (original == left) left = replacement; else if (original == right) right = replacement; else throw new IllegalArgumentException(this + " cannot replace a non-existant child");
    }
    
    /**
     * A primitive way to validate that nodes are stored in sorted
     * order and that their sizes are consistent. This throws a
     * IllegalStateException if any infraction is found.
     */
    void validate(IndexedTree host) {
        if (left != null) left.validate(host);
        if (right != null) right.validate(host);
        if (host.getComparator() != null && sorted) {
            if (leftSize > 0 && left.sorted && host.getComparator().compare(left.value, value) > 0) {
                throw new IllegalStateException("" + this + "left node, \"" + left + "\" larger than middle node, \"" + this + "\"");
            }
            if (rightSize > 0 && right.sorted && host.getComparator().compare(value, right.value) > 0) {
                throw new IllegalStateException("" + this + " middle node, \"" + this + "\" larger than right node, \"" + right + "\"");
            }
        }
        if ((left == null && leftSize != 0) || (left != null && leftSize != left.size())) {
            throw new IllegalStateException("Cached leftSize " + leftSize + " != reported left.size() " + left.size());
        }
        if ((right == null && rightSize != 0) || (right != null && rightSize != right.size())) {
            throw new IllegalStateException("Cached rightSize " + rightSize + " != reported right.size() " + right.size());
        }
    }
    
    /**
     * Ensures that the tree satisfies the AVL property.  It is sufficient to
     * recurse up the tree only as long as height recalculations are needed.
     * As such, this method is intended to be called only on a node whose height
     * may be out of sync due to an insertion or deletion.
     */
    private void ensureAVL(IndexedTree host) {
        int oldHeight = height;
        recalculateHeight();
        avlRotate(host);
        if (height != oldHeight && parent != null) parent.ensureAVL(host);
    }
    
    /**
     * Recalculates the cached height at this level.
     */
    private void recalculateHeight() {
        int leftHeight = left == null ? 0 : left.height;
        int rightHeight = right == null ? 0 : right.height;
        height = 1 + Math.max(leftHeight, rightHeight);
    }
    
    /**
     * Determines if AVL rotations are required and performs them if they are.
     */
    private void avlRotate(IndexedTree host) {
        int leftHeight = (left != null ? left.height : 0);
        int rightHeight = (right != null ? right.height : 0);
        if (leftHeight - rightHeight >= 2) {
            int leftLeftHeight = (left.left != null ? left.left.height : 0);
            int leftRightHeight = (left.right != null ? left.right.height : 0);
            if (leftRightHeight > leftLeftHeight) left.rotateRight(host);
            rotateLeft(host);
        } else if (rightHeight - leftHeight >= 2) {
            int rightLeftHeight = (right.left != null ? right.left.height : 0);
            int rightRightHeight = (right.right != null ? right.right.height : 0);
            if (rightLeftHeight > rightRightHeight) right.rotateLeft(host);
            rotateRight(host);
        }
    }
    
    /**
     * AVL-Rotates this subtree with its left child.
     *
     * For every link (left, right, parent), there are up to three
     * updates to be made. We need to set the new value on the
     * replacement, the new value on this, and the new value on the
     * other node.
     */
    private void rotateLeft(IndexedTree host) {
        IndexedTreeNode replacement = left;
        left = replacement.right;
        leftSize = replacement.rightSize;
        if (replacement.right != null) replacement.right.parent = this;
        replacement.right = this;
        replacement.rightSize = size();
        if (parent != null) parent.replaceChildNode(this, replacement); else host.setRootNode(replacement);
        replacement.parent = parent;
        parent = replacement;
        recalculateHeight();
        replacement.height = 0;
    }
    
    /**
     * AVL-Rotates this subtree with its right child.
     *
     * For every link (left, right, parent), there are up to three
     * updates to be made. We need to set the new value on the
     * replacement, the new value on this, and the new value on the
     * other node.
     */
    private void rotateRight(IndexedTree host) {
        IndexedTreeNode replacement = right;
        right = replacement.left;
        rightSize = replacement.leftSize;
        if (replacement.left != null) replacement.left.parent = this;
        replacement.left = this;
        replacement.leftSize = size();
        if (parent != null) parent.replaceChildNode(this, replacement); else host.setRootNode(replacement);
        replacement.parent = parent;
        parent = replacement;
        recalculateHeight();
        replacement.height = 0;
    }
    
    /**
     * Prints the tree by its contents.
     */
    public String toString() {
        String leftString = left == null ? "." : left.toString();
        String valueString = value instanceof IndexedTreeNode ? "NODE_" + ((IndexedTreeNode)(IndexedTreeNode)value).getIndex() : value.toString();
        String rightString = right == null ? "." : right.toString();
        return "(" + leftString + " " + valueString + " " + rightString + ")";
    }
}
