package ca.odell.glazedlists.io;

import ca.odell.glazedlists.*;
import ca.odell.glazedlists.event.*;
import ca.odell.glazedlists.impl.pmap.PersistentMap;
import ca.odell.glazedlists.impl.pmap.Chunk;
import ca.odell.glazedlists.impl.io.*;
import ca.odell.glazedlists.util.concurrent.*;
import java.util.*;
import java.io.*;

/**
 * An {@link EventList} that is persisted to disk.
 *
 * <p><font size="5"><strong><font color="#FF0000">Warning:</font></strong> This
 * class is a technology preview and is subject to API changes.</font>
 *
 * <p><table border="1" width="100%" cellpadding="3" cellspacing="0">
 * <tr class="tableheadingcolor"><td colspan=2><font size="+2"><b>EventList Overview</b></font></td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Writable:</b></td><td>yes</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Concurrency:</b></td><td>Requires {@link ReadWriteLock} for every access, even for single-threaded use</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Performance:</b></td><td>N/A</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Memory:</b></td><td>O(N)</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Unit Tests:</b></td><td>N/A</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Issues:</b></td><td>N/A</td></tr>
 * </table>
 *
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public final class FileList extends TransformedList {
    
    /**
     * the destination file, just for user convenience 
     */
    private File file = null;
    
    /**
     * how bytes are encoded and decoded 
     */
    private ByteCoder byteCoder;
    
    /**
     * the underlying storage of ListEvents 
     */
    private PersistentMap storage = null;
    
    /**
     * the ID of the next update to write to disc 
     */
    private int nextUpdateId = 81;
    
    /**
     * whether this list can be modified 
     */
    private boolean writable = true;
    
    /**
     * Create a {@link FileList} that stores its data in the specified file.
     */
    public FileList(File file, ByteCoder byteCoder) throws IOException {
        super(new BasicEventList());
        this.file = file;
        this.byteCoder = byteCoder;
        storage = new PersistentMap(file);
        SortedMap sequentialUpdates = new TreeMap();
        for (Iterator k = storage.keySet().iterator(); k.hasNext(); ) {
            Integer key = (Integer)(Integer)k.next();
            Bufferlo valueBuffer = ((Chunk)(Chunk)storage.get(key)).getValue();
            sequentialUpdates.put(key, valueBuffer);
        }
        for (Iterator u = sequentialUpdates.keySet().iterator(); u.hasNext(); ) {
            Integer key = (Integer)(Integer)u.next();
            Bufferlo update = (Bufferlo)(Bufferlo)sequentialUpdates.get(key);
            ListEventToBytes.toListEvent(update, this, byteCoder);
        }
        if (!sequentialUpdates.isEmpty()) {
            nextUpdateId = ((Integer)(Integer)sequentialUpdates.lastKey()).intValue() + 1;
        }
        source.addListEventListener(this);
    }
    
    /**
     * {@inheritDoc} 
     */
    public boolean isWritable() {
        return writable;
    }
    
    /**
     * {@inheritDoc} 
     */
    public void listChanged(ListEvent listChanges) {
        try {
            ListEvent listChangesCopy = new ListEvent(listChanges);
            Bufferlo listChangesBytes = ListEventToBytes.toBytes(listChangesCopy, byteCoder);
            storage.put(new Integer(nextUpdateId), new Chunk(listChangesBytes));
            nextUpdateId++;
        } catch (IOException e) {
            throw new IllegalStateException(e.getMessage());
        }
        updates.forwardEvent(listChanges);
    }
    
    /**
     * Closes this FileList so that it consumes no disc resources. The list may
     * continue to be read until it is {@link #dispose() disposed}.
     */
    public void close() {
        if (storage != null) storage.close();
        storage = null;
        writable = false;
    }
    
    /**
     * {@inheritDoc} 
     */
    public void dispose() {
        close();
        super.dispose();
    }
}
