package ca.odell.glazedlists.swing;

import ca.odell.glazedlists.*;
import ca.odell.glazedlists.gui.*;
import ca.odell.glazedlists.event.*;
import javax.swing.*;
import javax.swing.table.*;

/**
 * A {@link TableModel} that holds an {@link EventList}. Each element of the list
 * corresponds to a row in the {@link TableModel}. The columns of the table must
 * be specified using a {@link TableFormat}.
 *
 * <p>The EventTableModel class is <strong>not thread-safe</strong>. Unless otherwise
 * noted, all methods are only safe to be called from the event dispatch thread.
 * To do this programmatically, use {@link SwingUtilities#invokeAndWait(Runnable)}.
 *
 * @see <a href="http://publicobject.com/glazedlists/tutorial/">Glazed Lists Tutorial</a>
 *
 * @see SwingUtilities#invokeAndWait(Runnable)
 * @see <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=112">Bug 112</a>
 * @see <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=146">Bug 146</a>
 * @see <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=177">Bug 177</a>
 *
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public class EventTableModel extends AbstractTableModel implements ListEventListener {
    
    /**
     * the proxy moves events to the Swing Event Dispatch thread 
     */
    private TransformedList swingThreadSource = null;
    
    /**
     * Specifies how to render table headers and sort 
     */
    private TableFormat tableFormat;
    
    /**
     * Reusable table event for broadcasting changes 
     */
    private MutableTableModelEvent tableModelEvent = new MutableTableModelEvent(this);
    
    /**
     * Creates a new table that renders the specified list in the specified
     * format.
     */
    public EventTableModel(EventList source, TableFormat tableFormat) {
        super();
        this.swingThreadSource = GlazedListsSwing.swingThreadProxyList(source);
        this.tableFormat = tableFormat;
        swingThreadSource.addListEventListener(this);
    }
    
    /**
     * Creates a new table that renders the specified list with an automatically
     * generated {@link TableFormat}. It uses JavaBeans and reflection to create
     * a {@link TableFormat} as specified.
     *
     * <p>Note that the classes which will be obfuscated may not work with
     * reflection. In this case, implement a {@link TableFormat} manually.
     *
     * @param propertyNames an array of property names in the JavaBeans format.
     *      For example, if your list contains Objects with the methods getFirstName(),
     *      setFirstName(String), getAge(), setAge(Integer), then this array should
     *      contain the two strings "firstName" and "age". This format is specified
     *      by the JavaBeans {@link java.beans.PropertyDescriptor}.
     * @param columnLabels the corresponding column names for the listed property
     *      names. For example, if your columns are "firstName" and "age", then
     *      your labels might be "First Name" and "Age".
     * @param writable an array of booleans specifying which of the columns in
     *      your table are writable.
     */
    public EventTableModel(EventList source, String[] propertyNames, String[] columnLabels, boolean[] writable) {
        this(source, (TableFormat)GlazedLists.tableFormat(propertyNames, columnLabels, writable));
    }
    
    /**
     * Gets the Table Format.
     */
    public TableFormat getTableFormat() {
        return tableFormat;
    }
    
    /**
     * Sets this table to be rendered by a different table format. This has
     * some very important consequences. The selection will be lost - this is
     * due to the fact that the table formats may have different numbers of
     * columns. Another consequence is that the entire table will require
     * repainting. In a ScrollPane, only the currently displayed cells and
     * those above (before) them will require repainting. In order to provide
     * the best performance, the scroll pane may be scrolled to the top to
     * prevent a delay while rendering off-screen cells.
     */
    public void setTableFormat(TableFormat tableFormat) {
        this.tableFormat = tableFormat;
        tableModelEvent.setStructureChanged();
        fireTableChanged(tableModelEvent);
    }
    
    /**
     * Retrieves the value at the specified location from the table.
     *
     * <p>This may be used by renderers to paint the cells of a row differently
     * based on the entire value for that row. 
     *
     * @see ca.odell.glazedlists.swing.EventTableModel#getValueAt(int,int) ListTable
     */
    public Object getElementAt(int index) {
        swingThreadSource.getReadWriteLock().readLock().lock();
        try {
            return swingThreadSource.get(index);
        } finally {
            swingThreadSource.getReadWriteLock().readLock().unlock();
        }
    }
    
    /**
     * For implementing the ListEventListener interface. This sends changes
     * to the table which can repaint the table cells. Because this class uses
     * a EventThreadProxy, it is guaranteed that all natural
     * calls to this method use the Swing thread.
     *
     * <p>This tests the size of the change to determine how to handle it. If the
     * size of the change is greater than the changeSizeRepaintAllThreshhold,
     * then the entire table is notified as changed. Otherwise only the descrete
     * areas that changed are notified.
     */
    public void listChanged(ListEvent listChanges) {
        swingThreadSource.getReadWriteLock().readLock().lock();
        try {
            while (listChanges.nextBlock()) {
                int startIndex = listChanges.getBlockStartIndex();
                int endIndex = listChanges.getBlockEndIndex();
                int changeType = listChanges.getType();
                tableModelEvent.setValues(startIndex, endIndex, changeType);
                fireTableChanged(tableModelEvent);
            }
        } finally {
            swingThreadSource.getReadWriteLock().readLock().unlock();
        }
    }
    
    /**
     * Fetch the name for the specified column.
     */
    public String getColumnName(int column) {
        return tableFormat.getColumnName(column);
    }
    
    /**
     * The number of rows equals the number of entries in the source event list.
     */
    public int getRowCount() {
        swingThreadSource.getReadWriteLock().readLock().lock();
        try {
            return swingThreadSource.size();
        } finally {
            swingThreadSource.getReadWriteLock().readLock().unlock();
        }
    }
    
    /**
     * Get the column count as specified by the table format.
     */
    public int getColumnCount() {
        return tableFormat.getColumnCount();
    }
    
    /**
     * Gets the class of elements in the specified column. This behaviour can be
     * customized by implementing the {@link AdvancedTableFormat} interface.
     */
    public Class getColumnClass(int columnIndex) {
        if (tableFormat instanceof AdvancedTableFormat) {
            return ((AdvancedTableFormat)(AdvancedTableFormat)tableFormat).getColumnClass(columnIndex);
        } else {
            return super.getColumnClass(columnIndex);
        }
    }
    
    /**
     * Retrieves the value at the specified location from the table.
     *
     * <p>Before every get, we need to validate the row because there may be an
     * update waiting in the event queue. For example, it is possible that
     * the source list has been updated by a database thread. Such a change
     * may have been sent as notification, but after this request in the
     * event queue. In the case where a row is no longer available, null is
     * returned. The value returned is insignificant in this case because the
     * Event queue will very shortly be repainting (or removing) the row
     * anyway.
     */
    public Object getValueAt(int row, int column) {
        swingThreadSource.getReadWriteLock().readLock().lock();
        try {
            return tableFormat.getColumnValue(swingThreadSource.get(row), column);
        } finally {
            swingThreadSource.getReadWriteLock().readLock().unlock();
        }
    }
    
    /**
     * The list table is not editable. For an editable list table, use the
     * WritableListTable instead.
     */
    public boolean isCellEditable(int row, int column) {
        if (tableFormat instanceof WritableTableFormat) {
            WritableTableFormat writableTableFormat = (WritableTableFormat)(WritableTableFormat)tableFormat;
            swingThreadSource.getReadWriteLock().readLock().lock();
            try {
                Object toEdit = swingThreadSource.get(row);
                return writableTableFormat.isEditable(toEdit, column);
            } finally {
                swingThreadSource.getReadWriteLock().readLock().unlock();
            }
        } else {
            return false;
        }
    }
    
    /**
     * The list table is not editable. For an editable list table, use the
     * WritableListTable instead.
     */
    public void setValueAt(Object editedValue, int row, int column) {
        if (tableFormat instanceof WritableTableFormat) {
            swingThreadSource.getReadWriteLock().writeLock().lock();
            try {
                WritableTableFormat writableTableFormat = (WritableTableFormat)(WritableTableFormat)tableFormat;
                Object baseObject = swingThreadSource.get(row);
                Object updatedObject = writableTableFormat.setColumnValue(baseObject, editedValue, column);
                if (updatedObject != null) {
                    swingThreadSource.set(row, updatedObject);
                }
            } finally {
                swingThreadSource.getReadWriteLock().writeLock().unlock();
            }
        } else {
            throw new UnsupportedOperationException("Unexpected set() on read-only table");
        }
    }
    
    /**
     * Releases the resources consumed by this {@link EventTableModel} so that it
     * may eventually be garbage collected.
     *
     * <p>An {@link EventTableModel} will be garbage collected without a call to
     * {@link #dispose()}, but not before its source {@link EventList} is garbage
     * collected. By calling {@link #dispose()}, you allow the {@link EventTableModel}
     * to be garbage collected before its source {@link EventList}. This is 
     * necessary for situations where an {@link EventTableModel} is short-lived but
     * its source {@link EventList} is long-lived.
     * 
     * <p><strong><font color="#FF0000">Warning:</font></strong> It is an error
     * to call any method on a {@link EventTableModel} after it has been disposed.
     */
    public void dispose() {
        swingThreadSource.dispose();
    }
}
