package ca.odell.glazedlists;

import junit.framework.*;
import ca.odell.glazedlists.event.*;
import java.util.*;

/**
 * This test attempts to cause atomic change events that have change blocks
 * with indexes in random order.
 *
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public class IndexOrderTest extends TestCase {
    
    public IndexOrderTest() {
        super();
    }
    
    /**
     * for randomly choosing list indicies 
     */
    private Random random = new Random();
    
    /**
     * Test to verify that the list changes occur in increasing order.
     * GlazedL
     * <p>This creates a long chain of lists designed to cause events where the indicies
     * are out of order. The resultant list is a list of integer arrays of size two.
     * That list has been filtered to not contain any elements where the first index is
     * greater than 50. It has been sorted in increasing order.
     */
    public void testIncreasingOrder() {
        EventList unsorted = new BasicEventList();
        IntegerArrayMatcherEditor matcherEditor = new IntegerArrayMatcherEditor(0, 50);
        FilterList filteredOnce = new FilterList(unsorted, matcherEditor);
        for (int a = 0; a < 100; a++) {
            List currentChange = new ArrayList();
            for (int b = 0; b < 10; b++) {
                currentChange.add(new int[]{random.nextInt(100), random.nextInt(100)});
            }
            unsorted.addAll(currentChange);
        }
        for (int b = 0; b < 100; b++) {
            matcherEditor.setFilter(random.nextInt(2), random.nextInt(100));
        }
    }
    
    /**
     * Test to verify that the lists work with change indicies out of order.
     *
     * <p>This creates a long chain of lists designed to cause events where the indicies
     * are out of order. The resultant list is a list of integer arrays of size two.
     * That list has been filtered to not contain any elements where the first index is
     * greater than 50. It has been sorted in increasing order.
     */
    public void testIndexOutOfOrder() {
        EventList unsorted = new BasicEventList();
        SortedList sortedOnce = new SortedList(unsorted, GlazedListsTests.intArrayComparator(0));
        IntegerArrayMatcherEditor matcherEditor = new IntegerArrayMatcherEditor(0, 50);
        FilterList filteredOnce = new FilterList(sortedOnce, matcherEditor);
        SortedList sortedTwice = new SortedList(filteredOnce, GlazedListsTests.intArrayComparator(0));
        unsorted.addListEventListener(new IncreasingChangeIndexListener());
        sortedOnce.addListEventListener(new IncreasingChangeIndexListener());
        filteredOnce.addListEventListener(new IncreasingChangeIndexListener());
        ArrayList controlList = new ArrayList();
        for (int a = 0; a < 15; a++) {
            List currentChange = new ArrayList();
            for (int b = 0; b < controlList.size() || b < 10; b++) {
                currentChange.add(new int[]{random.nextInt(100), random.nextInt(100)});
            }
            unsorted.addAll(currentChange);
            controlList.addAll(currentChange);
            Collections.sort(controlList, sortedTwice.getComparator());
            for (Iterator i = controlList.iterator(); i.hasNext(); ) {
                if (matcherEditor.getMatcher().matches(i.next())) continue;
                i.remove();
            }
            assertEquals(controlList, filteredOnce);
        }
    }
    
    /**
     * A special list change listener that verifies that the change indicies
     * within each atomic change are in increasing order.
     */
    class IncreasingChangeIndexListener implements ListEventListener {
        
        IncreasingChangeIndexListener() {
            super();
        }
        
        public void listChanged(ListEvent listChanges) {
            StringBuffer changeDescription = new StringBuffer();
            int previousChangeIndex = -1;
            boolean increasingOrder = true;
            while (listChanges.next()) {
                int changeIndex = listChanges.getIndex();
                int changeType = listChanges.getType();
                if (changeType == ListEvent.UPDATE) {
                    changeDescription.append("U");
                } else if (changeType == ListEvent.INSERT) {
                    changeDescription.append("I");
                } else if (changeType == ListEvent.DELETE) {
                    changeDescription.append("D");
                }
                changeDescription.append(changeIndex);
                if (changeType == ListEvent.DELETE) {
                    if (changeIndex < previousChangeIndex) {
                        increasingOrder = false;
                        changeDescription.append("*");
                    }
                } else {
                    if (changeIndex <= previousChangeIndex) {
                        increasingOrder = false;
                        changeDescription.append("*");
                    }
                }
                changeDescription.append(" ");
                previousChangeIndex = changeIndex;
            }
            if (!increasingOrder) {
                fail("List changes not in increasing order: " + changeDescription);
            }
        }
    }
}
