package ca.odell.glazedlists;

import java.util.*;
import ca.odell.glazedlists.event.*;
import junit.framework.*;

/**
 * A very basic list that ensures that lists are kept consistent and that
 * the change events are consistent.
 *
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public class ListConsistencyListener implements ListEventListener {
    
    /**
     * a second copy of the list data 
     */
    private List expected;
    
    /**
     * a name for reporting problems with the list 
     */
    private String name;
    
    /**
     * the source list to compare against 
     */
    private EventList source;
    
    /**
     * whether to cough out changes to the console as they happen 
     */
    private boolean verbose = false;
    
    /**
     * Creates a new ListConsistencyListener that ensures events from the source
     * list are consistent.
     *
     * @param verbose whether to print changes to the console as they happne
     */
    public ListConsistencyListener(EventList source, String name, boolean verbose) {
        super();
        this.source = source;
        this.name = name;
        this.verbose = verbose;
        expected = new ArrayList();
        for (int i = 0; i < source.size(); i++) {
            expected.add(source.get(i));
        }
    }
    
    /**
     * Creates a new ListConsistencyListener that ensures events from the source
     * list are consistent.
     */
    public ListConsistencyListener(EventList source, String name) {
        this(source, name, false);
    }
    
    /**
     * Validate that this list is as expected.
     */
    public void assertConsistent() {
        Assert.assertEquals(expected, source);
    }
    
    /**
     * For implementing the ListEventListener interface.
     */
    public void listChanged(ListEvent listChanges) {
        Assert.assertEquals(source, listChanges.getSource());
        if (verbose) System.out.println(name + ": " + listChanges + ", size: " + source.size() + ", source: " + source);
        List changedIndices = new ArrayList();
        int highestChangeIndex = 0;
        if (listChanges.isReordering()) {
            int[] reorderMap = listChanges.getReorderMap();
            Assert.assertEquals(expected.size(), reorderMap.length);
            List newExpectedValues = new ArrayList(expected.size());
            for (int i = 0; i < reorderMap.length; i++) {
                newExpectedValues.add(i, expected.get(reorderMap[i]));
                changedIndices.add(new Integer(i));
            }
            this.expected = newExpectedValues;
        } else {
            while (listChanges.next()) {
                int changeIndex = listChanges.getIndex();
                int changeType = listChanges.getType();
                changedIndices.add(new Integer(changeIndex));
                Assert.assertTrue(changeIndex >= 0);
                Assert.assertTrue(changeIndex >= highestChangeIndex);
                highestChangeIndex = changeIndex;
                if (changeType == ListEvent.INSERT) {
                    expected.add(changeIndex, source.get(changeIndex));
                } else if (changeType == ListEvent.DELETE) {
                    expected.remove(changeIndex);
                } else if (changeType == ListEvent.UPDATE) {
                    expected.set(changeIndex, source.get(changeIndex));
                }
            }
        }
        Assert.assertEquals(expected.size(), source.size());
        for (Iterator c = changedIndices.iterator(); c.hasNext(); ) {
            int changeIndex = ((Integer)(Integer)c.next()).intValue();
            for (int i = Math.max(changeIndex - 1, 0); i < Math.min(changeIndex + 2, expected.size()); i++) {
                Assert.assertEquals(expected.get(i), source.get(i));
            }
        }
    }
}
